/*
 *  (c) 2005,2006 Frank A. Zdarsky, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.curves;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import unikl.disco.nc.Configuration;
import unikl.disco.misc.Rational;

/**
 * Class representing a piecewise linear curve, defined on [0,inf).<br/>
 * The curve is stored as an array of <code>LinearSegment</code>
 * objects. Each of these objects defines a linear piece of the
 * curve from one inflection point up to, but not including, the
 * next. It is possible to define discontinuities by defining two
 * subsequent <code>LinearSegment</code> instances which start at
 * the same inflection point. In this case, the second segment
 * needs to have <code>leftopen</code> set to <code>true</code> to
 * indicate that the inflection point is excluded from the second
 * segment.<br/>
 * All arithmetic operations on a curve return a new instance of class
 * <code>Curve</code>.<br/>
 *
 * @author Frank A. Zdarsky
 */
public class Curve {
	public static final Rational EPSILON = Rational.getEpsilon();

	private static final int OPERATOR_ADD = 0;
	private static final int OPERATOR_SUB = 1;
	private static final int OPERATOR_MIN = 2;
	private static final int OPERATOR_MAX = 3;

	private LinearSegment[] segments;
	private List<Curve> token_buckets;
	private List<Curve> rate_latencies;

	// TODO Implement tracing of the strictness property throughout DNClib  
	// I tried to use an explicit ServiceCurve class derived from this one, but that leads to a lot of duplicated code and loss of flexibility
	// TODO pseudo-constructors for service curves, arrival curves + checks and warnings during nc operations
//	private boolean is_service_curve = false;
//	private boolean is_strict_service_curve = false;
	// TODO does the current FIFO implementation demand a specific service curve?
//	private boolean is_fifo_service_curve = false;

	private String label;
	private String alias;
	
	private static final Curve NULL_CURVE = createHorizontal( new Rational( 0.0 ) );
	private static final Curve ZERO_DELAY_BURST_CURVE = createBurstDelay( new Rational( 0.0 ) );
	
	/**
	 * Creates a null curve.
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createNullCurve() {
		return NULL_CURVE.copy();
	}
	
	/**
	 * Creates a burst curve with zero delay.
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createZeroDelayBurst() {
		return ZERO_DELAY_BURST_CURVE.copy();
	}
	
	/**
	 * Creates a horizontal curve.
	 * @param y the y-intercept of the curve
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createHorizontal( Rational y ) {
		Curve c = new Curve(1);
		c.getSegment(0).y0 = y;
		return c;
	}

	/**
	 * Creates a horizontal curve.
	 * @param y the y-intercept of the curve
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createHorizontal( double y ) {
		return createHorizontal( new Rational( y ) );
	}

	/**
	 * Creates a burst delay curve.
	 * @param t the delay, which must be >= 0.0
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createBurstDelay( Rational t ) {
		Curve c = new Curve(2);

		c.getSegment(0).x0       = new Rational( 0.0 );
		c.getSegment(0).y0       = new Rational( 0.0 );
		c.getSegment(0).grad     = new Rational( 0.0 );
		c.getSegment(0).leftopen = false;
		
		c.getSegment(1).x0       = Rational.max( new Rational( 0.0 ), t);
		c.getSegment(1).y0       = Rational.getPosInfinity();
		c.getSegment(1).grad     = new Rational( 0.0 );
		c.getSegment(1).leftopen = true;
		
		if ( Configuration.createLabels() ) {
			c.setLabel( "\u03B4" + "(" + t.toString() + ")" );
		}
		return c;
	}
	
	/**
	 * Creates a burst delay curve.
	 * @param t the delay, which must be >= 0.0
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createBurstDelay( double t ) {
		return createBurstDelay( new Rational( t ) );
	}

	/**
	 * Creates a new token bucket curve.
	 * @param b the burstiness
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createTokenBucket( Rational b, Rational r ) {
		Curve c = new Curve(2);

		c.getSegment(0).x0       = new Rational( 0.0 );
		c.getSegment(0).y0       = new Rational( 0.0 );
		c.getSegment(0).grad     = new Rational( 0.0 );
		c.getSegment(0).leftopen = false;
		
		c.getSegment(1).x0       = new Rational( 0.0 );
		c.getSegment(1).y0       = b;
		c.getSegment(1).grad     = r;
		c.getSegment(1).leftopen = true;
		
		if ( Configuration.createLabels() ) {
			c.setLabel( "\u03B3" + "(" + b.toString() + "," + r.toString() + ")" );
		}
		return c;
	}

	/**
	 * Creates a new token bucket curve.
	 * @param b the burstiness
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createTokenBucket( double b, double r ) {
		return createTokenBucket( new Rational( b ), new Rational( r ) ) ;
	}

	/**
	 * Creates a new rate latency curve.
	 * @param t the latency
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createRateLatency( Rational t, Rational r ) {
		Curve c;
		if ( t.equals( 0.0 ) ) {
			c = new Curve(1);

			c.getSegment(0).x0       = new Rational( 0.0 );
			c.getSegment(0).y0       = new Rational( 0.0 );
			c.getSegment(0).grad     = r;
			c.getSegment(0).leftopen = false;
		} else {
			c = new Curve(2);
			
			c.getSegment(0).x0       = new Rational( 0.0 );
			c.getSegment(0).y0       = new Rational( 0.0 );
			c.getSegment(0).grad     = new Rational( 0.0 );
			c.getSegment(0).leftopen = false;

			c.getSegment(1).x0       = t;
			c.getSegment(1).y0       = new Rational( 0.0 );
			c.getSegment(1).grad     = r;
			c.getSegment(1).leftopen = false;
		}
		if ( Configuration.createLabels() ) {
			c.setLabel( "\u03B2" + "(" + t.toString() + "," + r.toString() + ")" );
		}
		return c;
	}

	/**
	 * Creates a new rate latency curve.
	 * @param t the latency
	 * @param r the rate
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createRateLatency( double t, double r ) {
		return createRateLatency( new Rational( t ), new Rational( r ) );
	}

	/**
	 * Creates a new curve from a list of token bucket curves.
	 * @param token_buckets a list of token bucket curves
	 * @return a <code>Curve</code> instance
	 */
	public static Curve createFromTokenBuckets(List<Curve> token_buckets) {
		Curve c = ZERO_DELAY_BURST_CURVE.copy();
		for (Iterator<Curve> tb_iter = token_buckets.iterator(); tb_iter.hasNext(); ) {
			c = min( c, (Curve) tb_iter.next() );
		}
		return c;
	}
	
	/**
	 * Creates an empty <code>Curve</code> instance.
	 */
	public Curve() {
		segments = new LinearSegment[0];
	}
	
	/**
	 * Creates a <code>Curve</code> instance with <code>segment_count</code>
	 * empty <code>LinearSegment</code> instances.
	 * @param segment_count the number of segments
	 */
	public Curve(int segment_count) {
		segments = new LinearSegment[segment_count];
		for (int i = 0; i < segment_count; i++) {
			segments[i] = LinearSegment.createHorizontalLine( 0.0 );
		}
	}

	/**
	 * Creates a <code>Curve</code> instance with inflection points at
	 * the x-coordinates <code>ip_xs</code> and with slopes defined
	 * by <code>ip_ms</code>. The y-coordinates are computed automatically.
	 * @param ip_xs
	 * @param ip_ms
	 */
	public Curve( Rational[] ip_xs, Rational[] ip_ms ) {
		this(ip_xs.length);
		for (int i = 0; i < segments.length; i++) {
			segments[i].x0   = ip_xs[i];
			segments[i].grad = ip_ms[i];
		}
		markDiscontinuities();
		computeYs(1);
	}

	/**
	 * Creates a <code>Curve</code> instance with inflection points at
	 * the x-coordinates <code>ip_xs</code> and with slopes defined
	 * by <code>ip_ms</code>. This constructor accepts an array of
	 * y-coordinates so that discontinuities can be created. The
	 * parameter <code>compute_y_start</code> indicates from which
	 * inflection point (starting with 0), the y-coordinate is again
	 * computed automatically. Note that all arrays need to have the
	 * same length.
	 * @param ip_xs
	 * @param ip_ms
	 */
	public Curve( Rational[] ip_xs, Rational[] ip_ys, Rational[] ip_ms,
			int compute_y_start) {
		this(ip_xs.length);
		for (int i = 0; i < segments.length; i++) {
			segments[i].x0   = ip_xs[i];
			segments[i].y0   = ip_ys[i];
			segments[i].grad = ip_ms[i];
		}
		markDiscontinuities();
		if (compute_y_start >= 1) {
			computeYs(compute_y_start);
		}
	}
	
	/**
	 * Returns a copy of this instance.
	 * @return a copy of this instance.
	 */
	public Curve copy() {
		Curve copy = new Curve(segments.length);
		for (int i = 0; i < segments.length; i++) {
			copy.segments[i] = segments[i].copy();
		}
		copy.rate_latencies = null;
		copy.token_buckets = null;
		copy.label = this.label; // can copy directly as strings are unmutable
		copy.alias = this.alias;
		return copy;
	}

	public String getLabel() {
		return (label != null) ? label : "";
	}
	
	public void setLabel( String lab ) {
		label = lab;
	}

	public void setAlias(String alias) {
		this.alias = alias;
	}

	public String getAlias() {
		return (alias != null) ? alias : getLabel();
	}

	/**
	 * Returns the number of segments in this curve.
	 * @return the number of segments
	 */
	public int getSegmentCount() {
		return segments.length;
	}

	/**
	 * Returns the x-coordinate of inflection point <code>i</code>.
	 * @param i the index of the IP.
	 * @return the x-coordinate of the inflection point.
	 */
	public Rational getIPX(int i) {
		return segments[i].x0;
	}

	/**
	 * Returns the y-coordinate of inflection point <code>i</code>.
	 * @param i the index of the IP.
	 * @return the y-coordinate of the inflection point.
	 */
	public Rational getIPY(int i) {
		return segments[i].y0;
	}

	/**
	 * Returns the sustained rate (the gradient of the last segment).
	 * @return the sustained rate.
	 */
	public Rational getSustainedRate() {
		return segments[segments.length-1].grad;
	}

	/**
	 * Returns the highest rate (the steepest gradient of any segment).
	 * @return the highest rate.
	 */
	public Rational getHighestRate() {
		Rational max_rate = Rational.getNegInfinity();
		for (int i = 0; i < segments.length; i++) {
			if ( !isDiscontinuity(i) && ( segments[i].grad ).greater( max_rate ) ) {
				max_rate = segments[i].grad;
			}
		}
		return max_rate;
	}

	/**
	 * Sets the sustained rate (the gradient of the last segment).
	 * Note: No checks are performed when setting the sustained rate.
	 * The caller has to ensure that <code>r</code> is larger (smaller) than
	 * the gradient of the 2nd-to-last segment to keep the curve convex (concave).
	 * @param r the new sustained rate
	 */
	public void setSustainedRate( Rational r ) {
		segments[segments.length-1].grad = r;
	}

	/**
	 * Returns whether the inflection point <code>i</code> is excluded
	 * from the segment or not.
	 * @param i the index of the IP.
	 * @return <code>true</code> if the IP is excluded, <code>false</code> if not.
	 */
	public boolean isLeftopen(int i) {
		return segments[i].leftopen;
	}

	/**
	 * Returns whether the inflection point is a (real or unreal) discontinuity.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is a discontinuity, <code>false</code> if not.
	 */
	public boolean isDiscontinuity(int i) {
		return (i+1 < segments.length && ( Rational.abs( Rational.sub( segments[i+1].x0, segments[i].x0 ) ) ).less( EPSILON ) );
	}

	/**
	 * Returns whether the inflection point is a real discontinuity, i.e. the y0
	 * of the leftopen segment differs from the previous one.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is a real discontinuity, <code>false</code> if not.
	 */
	public boolean isRealDiscontinuity(int i) {
		return (isDiscontinuity(i) && ( Rational.abs( Rational.sub( segments[i+1].y0, segments[i].y0 ) ) ).ge( EPSILON ) );
	}

	/**
	 * Returns whether the inflection point is an unreal discontinuity, i.e. the y0
	 * of the leftopen segment is coincident with the y0 of the previous segment
	 * and therefore the unreal discontinuity may safely be removed.
	 * @param i the index of the IP
	 * @return <code>true</code> if the IP is an unreal discontinuity, <code>false</code> if not.
	 */
	public boolean isUnrealDiscontinuity(int i) {
		return (isDiscontinuity(i) && ( Rational.abs( Rational.sub(segments[i+1].y0, segments[i].y0 ) ) ).less( EPSILON ) );
	}

	/**
	 * Returns whether the current curve is a burst delay curve.
	 * @return <code>true</code> if the curve is burst delay, <code>false</code> otherwise
	 */
	public boolean isBurstDelay() {
		return (segments.length == 2
				&& segments[0].x0.equals( 0.0 ) && segments[0].y0.equals( 0.0 )
				&& segments[0].grad.equals( 0.0 ) && !segments[0].leftopen
				&& segments[1].x0.ge( new Rational( 0.0 ) ) && segments[1].y0.equals( Rational.getPosInfinity() )
				&& segments[1].grad.equals( 0.0 ) && segments[1].leftopen);
	}

	/**
	 * Adds a <code>LinearSegment</code> to the end of the curve.<br/>
	 * Note: It is the user's responsibility to add segments in the
	 * order of increasing x-coordinates.
	 * @param s the segment to be added.
	 */
	public void addSegment(LinearSegment s) {
		addSegment(segments.length, s);
	}

	/**
	 * Adds a <code>LinearSegment</code> at the location <code>pos</code>
	 * of the curve.<br/>
	 * Note: It is the user's responsibility to add segments in the
	 * order of increasing x-coordinates.
	 * @param pos the index into the segment array to add the new segment.
	 * @param s the segment to be added.
	 */
	public void addSegment(int pos, LinearSegment s) {
		if (pos < 0 || pos > segments.length) {
			throw new IllegalArgumentException("Index out of bounds (pos=" + pos + ")!");
		}
		if (s == null) {
			throw new IllegalArgumentException("Tried to insert null!");
		}
		LinearSegment[] old_segments = segments;
		segments = new LinearSegment[old_segments.length + 1];
		segments[pos] = s;
		if (pos > 0) {
			System.arraycopy(old_segments, 0, segments, 0, pos);
		}
		if (pos < old_segments.length) {
			System.arraycopy(old_segments, pos, segments, pos+1, old_segments.length-pos);
		}
	}

	/**
	 * Removes the segment at position <code>pos</code>.
	 * @param pos the index of the segment to be removed.
	 */
	public void removeSegment(int pos) {
		if (pos < 0 || pos >= segments.length) {
			throw new IllegalArgumentException("Index out of bounds (pos=" + pos + ")!");
		}
		LinearSegment[] old_segments = segments;
		segments = new LinearSegment[old_segments.length - 1];
		System.arraycopy(old_segments, 0, segments, 0, pos);
		System.arraycopy(old_segments, pos+1, segments, pos, old_segments.length-pos-1);
	}

	/**
	 * Marks discontinuities (subsequent segments having the same
	 * x-coordinate) by setting <code>leftopen</code> of the second segment.
	 */
	public void markDiscontinuities() {
		for (int i = 0; i < segments.length; i++) {
			segments[i].leftopen = false;
		}
		for (int i = 1; i < segments.length; i++) {
			if (segments[i-1].x0 == segments[i].x0) {
				segments[i].leftopen = true;
			}
		}
	}

	/**
	 * Computes the y-coordinates of inflection points starting with
	 * inflection point <code>start</code>.
	 * @param start the IP at which to start computing y-coordinates.
	 */
	public void computeYs(int start) {
		if (start < 1) {
			throw new IllegalArgumentException("Value of 'start' must be >= 1!");
		}
		for (int i = start; i < segments.length; i++) {
			Rational dx = Rational.sub( segments[i].x0, segments[i-1].x0 );
			segments[i].y0 = Rational.add( segments[i-1].y0, Rational.mult( dx, segments[i-1].grad ) );
		}
	}

	/**
	 * Decomposes this curve into a list of token bucket curves and
	 * stores this list in the curve's <code>token_buckets</code> field.<br/>
	 * Note: Curve must be concave.
	 */
	private void decomposeIntoTokenBuckets() {
		if( Configuration.performChecks() && !this.isConcave() ) {
			throw new RuntimeException("Can only decompose concave functions into token buckets.");
		}

		token_buckets = new ArrayList<Curve>();
		for (int i = 0; i < segments.length; i++) {
			if (isDiscontinuity(i)) {
				continue;
			}
			Rational b = Rational.sub( segments[i].y0, Rational.mult( segments[i].x0, segments[i].grad ) );
			Rational r = segments[i].grad;
			token_buckets.add( createTokenBucket(b, r) );
		}
	}

	/**
	 * Returns the number of token buckets the curve can be decomposed into.
	 * @return the number of token buckets
	 */
	public int getTBComponentCount() {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets.size();
	}

	/**
	 * Returns a list of token bucket curves that this curve can be
	 * decomposed into.
	 * @return the list of token buckets
	 */
	public final List<Curve> getTBComponents() {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets;
	}

	/**
	 * Returns the <code>i</code>th token bucket curve that this curve can be
	 * decomposed into.
	 * @param i the number of the token bucket
	 * @return the token bucket
	 */
	public Curve getTBComponent(int i) {
		if (token_buckets == null) {
			decomposeIntoTokenBuckets();
		}
		return (Curve) token_buckets.get(i);
	}

	/**
	 * Returns the burstiness of this token bucket curve.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 * @return the burstiness
	 */
	public Rational getTBBurst() {
		return segments[1].y0;
	}

	/**
	 * Sets the burstiness of this token bucket curve.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 * @param b the burstiness
	 */
	public void setTBBurst( Rational b ) {
		segments[1].y0 = b;
	}

	/**
	 * Makes the token bucket curve right-continuous.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 */
	public void makeTBRightContinuous() {
		segments[0].y0 = segments[1].y0;
	}

	/**
	 * Undoes the makeTBRightContinuous() operation.<br/>
	 * Note: For performance reasons there are no sanity checks! Only
	 *       call this method on a valid token bucket curve!
	 */
	public void undoMakeTBRightContinuous() {
		segments[0].y0 = new Rational( 0.0 );
	}

	/**
	 * Decomposes this curve into a list of rate latency curves and
	 * stores this list in the curve's <code>rate_latencies</code> field.<br/>
	 * Note: Curve must be convex.
	 */
	private void decomposeIntoRateLatencies() {
		if( Configuration.performChecks() && !this.isConvex() ) {
			throw new RuntimeException("Can only decompose convex functions into rate latency curves.");
		}

		rate_latencies = new ArrayList<Curve>();
		for (int i = 0; i < segments.length; i++) {
			if (segments[i].y0.equals( 0.0 ) && segments[i].grad.equals( 0.0 ) ) {
				continue;
			}
			Rational l = Rational.sub( segments[i].x0, Rational.div( segments[i].y0, segments[i].grad ) );
			Rational r = segments[i].grad;
			rate_latencies.add( createRateLatency(l, r) );
		}
	}

	/**
	 * Returns the number of rate latency curves the curve can be decomposed into.
	 * @return the number of rate latency curves
	 */
	public int getRLComponentCount() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies.size();
	}

	/**
	 * Returns a list of rate latency curves that this curve can be
	 * decomposed into.
	 * @return the list of rate latency curves
	 */
	public final List<Curve> getRLComponents() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies;
	}

	/**
	 * Returns the <code>i</code>th rate latency curve that this curve can be
	 * decomposed into.
	 * @param i the number of the rate latency curve
	 * @return the rate latency curve
	 */
	public Curve getRLComponent(int i) {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return (Curve) rate_latencies.get(i);
	}

	/**
	 * Removes unnecessary segments.
	 */
	public void beautify() {
		int i = 0;
		while( i < segments.length-1 ) {
			// remove unreal discontinuity
			if ( isUnrealDiscontinuity( i ) ) {
				segments[i+1].leftopen = segments[i].leftopen;
				removeSegment(i);
				continue;
			}
			i++;
		}

		i = 0;
		while(i < segments.length-1) {
			// join colinear segments
			Rational firstArg = Rational.sub( segments[i+1].grad, segments[i].grad );
			
			Rational secondArg = Rational.sub( segments[i+1].x0, segments[i].x0 );
			secondArg = Rational.mult( secondArg, segments[i].grad );
			secondArg = Rational.add( segments[i].y0, secondArg);
			secondArg = Rational.sub( segments[i+1].y0, secondArg );
			
			if ( Rational.abs( firstArg ).less( EPSILON )				// 2nd iteration yields a wrong true!
					&& Rational.abs( secondArg ).less( EPSILON ) ){
				
				removeSegment(i+1);
				if (i+1 < segments.length && !segments[i+1].leftopen) {
					Rational resultPt1 = Rational.sub( segments[i+1].y0, segments[i].y0 );
					Rational resultPt2 = Rational.sub( segments[i+1].x0, segments[i].x0 );
					
					segments[i].grad = Rational.div( resultPt1, resultPt2 );
				}
				continue;
			}
			i++;
		}

		for (i = 0; i < segments.length-1; i++) {
			if (segments[i].x0.equals( segments[i+1].x0 ) ) {
				segments[i].grad = new Rational( 0.0 );
			}
		}
	}

	/**
	 * Returns the number of the segment that defines the function
	 * value at x-coordinate <code>x</code>. The
	 * number of the segment is usually the same as the one
	 * returned by <code>getSegmentLimitRight(x)</code>, except for
	 * if a segment starts at <code>x</code> and is left-open.
	 * In this case the function returns the previous segment,
	 * rather than the current segment, as the previous segment
	 * defines <code>x</code>.
	 * @param x the x-coordinate
	 * @return the index of the segment into the array.
	 */
	public int getSegmentDefining( Rational x ) {
		for (int i = segments.length - 1; i >= 0; i--) {
			if (segments[i].leftopen) {
				if ( segments[i].x0.less( x ) ) {
					return i;
				}
			} else {
				if ( segments[i].x0.le( x ) ) {
					return i;
				}
			}
		}
		return -1;
	}
	
	public LinearSegment[] getSegments() {
		return segments;
	}
	
	public LinearSegment getSegment(int i) {
		if(i>=0) {
			return segments[i];
		} else {
			return segments[segments.length + i];
		}
	}

	/**
	 * Returns the function value at x-coordinate <code>x</code>, if
	 * <code>x>=0</code>, and <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public Rational f( Rational x ) {
		int i = getSegmentDefining(x);
		if (i < 0) {
			return Rational.getNaN();
		}
		return Rational.add( Rational.mult( Rational.sub( x, segments[i].x0 ), segments[i].grad ), segments[i].y0 );
	}

	/**
	 * Returns the curve's gradient at x-coordinate <code>x</code>, if
	 * <code>x>=0</code>, and <code>NaN</code> if not. Note that the
	 * gradient returned at discontinuities is <code>0.0</code>, the
	 * gradient returned at an inflection point is the gradient of the
	 * linear segment right of the inflection point, but only if the
	 * segment is not left-open.
	 * @param x the x-coordinate
	 * @return the gradient
	 */
	public Rational getGradientAt( Rational x ) {
		int i = getSegmentDefining(x);
		if (i < 0) {
			return Rational.getNaN();
		}
		return segments[i].grad;
	}

	/**
	 * Returns the number of the segment that defines the value
	 * of the function when computing the limit to the right
	 * of the function at x-coordinate <code>x</code>. The
	 * number of the segment is usually the same as the one
	 * returned by <code>getSegmentDefining(x)</code>, except for
	 * if a segment starts at <code>x</code> and is left-open.
	 * In this case the function returns the current segment,
	 * rather than the previous segment.
	 * @param x the x-coordinate
	 * @return the index of the segment into the array.
	 */
	public int getSegmentLimitRight( Rational x ) {
		for (int i = segments.length - 1; i >= 0; i--) {
			if ( segments[i].x0.le( x ) ) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns the limit to the right of the function value at
	 * x-coordinate <code>x</code>, if <code>x>=0</code>, and
	 * <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public Rational fLimitRight( Rational x ) {
		int i = getSegmentLimitRight(x);
		if (i < 0) {
			return Rational.getNaN();
		}
		return Rational.add( Rational.mult( Rational.sub( x, segments[i].x0 ), segments[i].grad ), segments[i].y0 );
	}

	/**
	 * Returns the gradient to the right of the function value at
	 * x-coordinate <code>x</code>, if <code>x>=0</code>, and
	 * <code>NaN</code> if not.
	 * @param x the x-coordinate
	 * @return the function value
	 */
	public Rational getGradientLimitRight( Rational x ) {
		int i = getSegmentLimitRight(x);
		if (i < 0) {
			return Rational.getNaN();
		}
		return segments[i].grad;
	}

	/**
	 * Returns the first segment at which the function reaches the
	 * value <code>y</code>. It returns -1 if the curve never
	 * reaches this value.
	 * @param y the y-coordinate
	 * @return the segment number
	 */
	public int getSegmentFirstAtValue( Rational y ) {
		if ( segments.length == 0 || segments[0].y0.greater( y ) ) {
			return -1;
		}
		for (int i = 0; i < segments.length; i++) {
			if (i < segments.length-1) {
				if ( segments[i+1].y0.ge( y ) ) {
					return i;
				}
			} else {
				if ( segments[i].grad.greater( new Rational( 0.0 ) ) ) {
					return i;
				}
			}
		}
		return -1;
	}

	/**
	 * Returns the smallest x value at which the function value is
	 * equal to <code>y</code>.
	 * @param y the y-coordinate
	 * @return the smallest x value
	 */
	public Rational f_inv( Rational y ) {
		return f_inv(y, false);
	}

	/**
	 * Returns the x value at which the function value is
	 * equal to <code>y</code>. If <code>rightmost</code> is
	 * <code>true</code>, returns the rightmost x-coordinate,
	 * otherwise the leftmost coordinate.
	 * @param y the y-coordinate
	 * @return the smallest x value
	 */
	public Rational f_inv( Rational y, boolean rightmost) {
		int i = getSegmentFirstAtValue(y);
		if (i < 0) {
			return Rational.getNaN();
		}
		if (rightmost) {
			while(i < segments.length && segments[i].grad.equals( new Rational( 0.0 ) ) ) {
				i++;
			}
			if (i >= segments.length) {
				return Rational.getPosInfinity();
			}
		}
		if ( !segments[i].grad.equals( new Rational( 0.0 ) ) ) {
			return Rational.add( segments[i].x0, Rational.div( Rational.sub( y, segments[i].y0 ), segments[i].grad ) );
		} else {
			return segments[i].x0;
		}
	}

	/**
	 * Returns the smallest x for which the curve is defined.
	 * @return the lower bound of the definition interval.
	 */
	public double getMinX() {
		return 0.0;
	}

	/**
	 * Returns the largest x for which the curve is defined.
	 * @return the upper bound of the definition interval.
	 */
	public Rational getMaxX() {
		return Rational.getPosInfinity();
	}

	public Rational getMaxY( Rational a, Rational b ) {
		int sa = getSegmentDefining(a);
		int sb = getSegmentDefining(b);

		Rational ret = new Rational( 0.0 );
		for(int i=sa; i<=sb; ++i) {
			Rational end = ( i+1 < segments.length ? Rational.min(getSegment(i+1).x0, b) : b );
			ret = Rational.max(ret, getSegment(i).maxAfter(a, end));
		}

		return ret;
	}

	/**
	 * Tests whether the curve is wide-sense increasing.
	 * @return whether the curve is wide-sense increasing.
	 */
	public boolean isWideSenseIncreasing() {
		Rational y = Rational.getNegInfinity();
		for (int i = 0; i < segments.length; i++) {
			if ( segments[i].y0.less( y ) || segments[i].grad.less( new Rational( 0.0 ) ) ) {
				return false;
			}
			y = segments[i].y0;
		}
		return true;
	}

	/**
	 * Tests whether the curve is wide-sense increasing and satisfies f(0)=0.
	 * @return whether the curve is an arrival curve.
	 */
	public boolean isArrivalCurve() {
		return ( isWideSenseIncreasing() && segments[0].y0.equals( new Rational( 0.0 ) ) );
	}
	
	/**
	 * Returns a copy of this curve ensuring that f(0)=0.
	 * @return a copy with with f(0) set to 0
	 */
	public static Curve makeArrivalCurve( Curve curve ) {
		if (curve.getSegment(0).y0.greater( new Rational( 0.0 ) ) ) {
			Curve result = new Curve(curve.getSegmentCount()+1);
			System.arraycopy(curve.getSegments(), 0, result.getSegments(), 1, curve.getSegmentCount());
			result.getSegment(1).leftopen = true;
			result.setLabel ( "" + curve.getLabel() );
			return result;
		} else {
			return curve.copy();
		}
	}

	/**
	 * Tests whether the curve is convex.
	 * @return whether the curve is convex.
	 */
	public boolean isConvex() {
		return isConvexIn( new Rational( 0.0 ), Rational.getPosInfinity() );
	}

	/**
	 * Tests whether the curve is convex in [a,b].
	 * @param a the lower bound of the test interval.
	 * @param b the upper bound of the test interval.
	 * @return whether the curve is convex
	 */
	public boolean isConvexIn( Rational a, Rational b ) {
		Rational last_gradient = Rational.getNegInfinity();

		int i_start = getSegmentDefining(a);
		int i_end   = getSegmentDefining(b);
		for (int i = i_start; i <= i_end; i++) {
			if (i_start < 0) {
				break;
			}
			if (i == i_end && segments[i].x0 == b) {
				break;
			}
			Rational gradient;
			if (i < segments.length-1) {
				gradient = Rational.div( Rational.sub( segments[i+1].y0, segments[i].y0 ),
											Rational.sub( segments[i+1].x0, segments[i].x0 ) );
			} else {
				gradient = segments[i].grad;
			}
			if ( gradient.less( last_gradient ) ) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}

	/**
	 * Tests whether the curve is concave.
	 * @return whether the curve is concave.
	 */
	public boolean isConcave() {
		return isConcaveIn( new Rational( 0.0 ), Rational.getPosInfinity() );
	}

	/**
	 * Tests whether the curve is concave in [a,b].
	 * @param a the lower bound of the test interval.
	 * @param b the upper bound of the test interval.
	 * @return whether the curve is concave.
	 */
	public boolean isConcaveIn( Rational a, Rational b ) {
		Rational last_gradient = Rational.getPosInfinity();

		int i_start = getSegmentDefining( a );
		int i_end   = getSegmentDefining( b );
		for ( int i = i_start; i <= i_end; i++ ) {
			if ( i == i_end && segments[i].x0 == b ) {
				break;
			}
			Rational gradient;
			// handles discontinuities
			if ( i < segments.length-1 ) {
				gradient = Rational.div( Rational.sub( segments[i+1].y0, segments[i].y0 ),
											Rational.sub( segments[i+1].x0, segments[i].x0 ) );
			} else {
				gradient = segments[i].grad;
			}
			if ( gradient.greater( last_gradient ) ) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}

	/**
	 * Tests whether the curve is almost concave, i.e. it is concave once its
	 * function value is larger than 0.
	 * @return whether the curve is almost concave.
	 */
	public boolean isAlmostConcave() {
		Rational last_gradient = Rational.getPosInfinity();

		for (int i = 0; i < segments.length; i++) {
			// skip the horizontal part at the beginning
			if ( last_gradient.equals( Rational.getPosInfinity() ) && segments[i].grad.equals( new Rational( 0.0 ) ) ) {
				continue;
			}

			Rational gradient;
			if (i < segments.length-1) {
				gradient = Rational.div( Rational.sub( segments[i+1].y0, segments[i].y0 ),
											Rational.sub( segments[i+1].x0, segments[i].x0 ) );
			} else {
				gradient = segments[i].grad;
			}
			if ( gradient.greater( last_gradient ) ) {
				return false;
			}
			last_gradient = gradient;
		}
		return true;
	}
	
	/**
	 * Returns a copy of this curve that is shifted to the right by <code>dx</code>,
	 * i.e. g(x) = f(x-dx).
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve shiftRight( Rational dx ) {
		if ( !( ( segments[0].y0 ).equals( 0.0 ) ) ) {
			throw new RuntimeException("Curve to shift right must pass through origin!");
		}
		Curve result = new Curve(segments.length + 1);
		result.segments[0] = LinearSegment.createHorizontalLine( 0.0 );
		result.segments[0].y0 = this.segments[0].y0; // decide what to do if pass thru origin req. dropped...
		for (int i = 0; i < segments.length; i++) {
			result.segments[i+1] = segments[i].copy();
			result.segments[i+1].x0 = Rational.add( result.segments[i+1].x0, dx );
		}
		result.beautify();
		if ( Configuration.createLabels() ) {
			result.label = getAlias() + "(t-" + dx.toString() + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve that is shifted to the left by <code>dx</code>,
	 * i.e. g(x) = f(x+dx). Note that the new curve is clipped at the y-axis so
	 * that in most cases <code>c.shiftLeftClipping(dx).shiftRight(dx) != c</code>!
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve shiftLeftClipping( Rational dx ) {
		Curve result = this.copy();
		int i = result.getSegmentDefining(dx);
		if ( result.segments[i].x0.less( dx ) ) {
			result.segments[i].y0 = Rational.add( result.segments[i].y0, 
									Rational.mult( Rational.sub( dx, result.segments[i].x0 ), result.segments[i].grad ) );
			result.segments[i].x0 = dx;
			result.segments[i].leftopen = false;
		}
		if (i > 0) {
			LinearSegment[] old_segments = result.segments;
			result.segments = new LinearSegment[old_segments.length - i];
			System.arraycopy(old_segments, i, result.segments, 0, result.segments.length);
		}
		for (i = 0; i < result.segments.length; i++) {
			result.segments[i].x0 = Rational.sub( result.segments[i].x0, dx );
		}
		result.beautify();

		if ( Configuration.createLabels() ) {
			result.label = getAlias() + "(t+" + dx.toString() + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve with latency removed, i.e. shifted left by
	 * the latency.
	 * @return a copy of this curve without latency
	 */
	public Curve removeLatency() {
		Curve result = this.copy();

		// remove all segment(s) with y0==0.0 and grad==0.0
		while(result.segments.length > 0) {
			if ( result.segments[0].y0.greater( new Rational( 0.0 ) ) || result.segments[0].grad.greater( new Rational( 0.0 ) ) ) {
				break;
			}
			if ( segments[0].y0.less( new Rational( 0.0 ) ) || segments[0].grad.less( new Rational ( 0.0 ) ) ) {
				throw new RuntimeException("Should have avoided neg. gradients elsewhere...");
			}
			result.removeSegment(0);
		}

		// in case that we've removed everything, the curve had infinite
		// latency, so return the NULL curve.
		if (result.segments.length == 0) {
			return createNullCurve();
		}

		// shift remaining segments left by latency
		Rational L = result.segments[0].x0;
		for (int i = 0; i < result.segments.length; i++) {
			result.segments[i].x0 = Rational.sub( result.segments[i].x0, L );
		}
		if (result.segments[0].leftopen) {
			result.addSegment(0, LinearSegment.createHorizontalLine( 0.0 ));
		}
		result.beautify();
		if ( Configuration.createLabels() ) {
			result.label = getAlias() + "(t+" + L + ")";
		}
		return result;
	}

	/**
	 * Returns a copy of this curve shifted vertically by <code>dy</code>.
	 * @param dy the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public Curve add( Rational dy ) {
		Curve result = this.copy();
		for (int i = 0; i < segments.length; i++) {
			result.segments[i].y0 = Rational.add( result.segments[i].y0, dy );
		}
		if ( Configuration.createLabels() ) {
			result.label = "(" + getAlias() + (( dy.ge( new Rational( 0.0 ) ) ) ? "+" : "") + dy.toString() + ")";
		}
		return result;
	}

	/**
	 * Returns the x-coordinate of the inflection point after which the function
	 * values are greater than zero.
	 * @return the latency of this curve.
	 */
	public Rational getLatency() {
		if ( segments[0].y0.greater( new Rational( 0.0 ) ) ) {
			return new Rational( 0.0 );
		}
		for (int i = 0; i < segments.length; i++) {
			Rational y0 = segments[i].y0;
			if (y0.less( new Rational( 0.0 ) ) && y0.greater( Rational.negate( EPSILON ) ) ) {
				y0 = new Rational( 0.0 );
			}
			if ( y0.greater( new Rational( 0.0 ) )
					|| ( y0.ge( new Rational( 0.0 ) ) && segments[i].grad.greater( new Rational( 0.0 ) ) )
				) {
				return segments[i].x0;
			}
			if (y0.less( new Rational( 0.0 ) ) || segments[i].grad.less( new Rational( 0.0 ) ) ) {
				throw new RuntimeException("Should have avoided neg. gradients elsewhere...");
			}
		}
		return Rational.getPosInfinity();
	}

	/**
	 * Returns a list containing each segment's height when projected onto
	 * the y-axis.
	 * @return a list of doubles containing the segment heights
	 */
	public List<Rational> getYIntervals() {
		List<Rational> y_intervals = new ArrayList<Rational>();
		for (int i = 1; i < segments.length; i++) {
			if (isDiscontinuity(i-1)) {
				continue;
			}
			y_intervals.add( Rational.sub( segments[i].y0, segments[i-1].y0 ) );
		}
		return y_intervals;
	}

	/**
	 * Approximates arrival curve by removing all segments whose projection onto
	 * the y-axis has a height smaller than <code>dy</code>. Returns the number of
	 * removed segments.
	 * @param dy the height below which segments are removed
	 * @return the number of removed segments
	 */
	public int approximateSegments( Rational dy ) {
		int segments_removed = 0;
		for (int i = segments.length-1; i > 0; i--) {
			if (isDiscontinuity(i-1)) {
				continue;
			}
			if ( Rational.sub( segments[i].y0, segments[i-1].y0 ).le( dy ) ) {
				approximateAtSegment(i-1);
				segments_removed++;
				break;
			}
		}
		return segments_removed;
	}

	/**
	 * Approximates a (concave) arrival curve by removing the specified segment.
	 * @param i the index of the segment to remove
	 */
	public void approximateAtSegment(int i) {
		if( Configuration.performChecks() && !isConcave() ) {
			throw new IllegalArgumentException("Curve to approximate must be concave!");
		}

		if (segments.length <= 2 || i < 1 || i >= segments.length-1) {
			return;
		}

		removeSegment(i);

		Rational x_cross;
		if (i == 1) {
			x_cross = new Rational( 0.0 );
		} else {
			x_cross = segments[i-1].getXIntersectionWith(segments[i]);
		}
		segments[i].y0       = Rational.sub( segments[i].y0, Rational.mult( Rational.sub( segments[i].x0, x_cross ), segments[i].grad ) );
		segments[i].x0       = x_cross;
		segments[i].leftopen = (i == 1);
	}
	
	// TODO rewrite
	public static Rational getXIntersection( Curve curve1, Curve curve2 ){
		Rational x_int = Rational.getPosInfinity();
		
		// inefficient
		for( int i = 0; i < curve1.getSegmentCount(); i++ ){
			boolean curve1_last = (i == curve1.getSegmentCount()-1);
					
			for( int j = 0; j < curve2.getSegmentCount(); j++ ){
				boolean curve2_last = (j == curve2.getSegmentCount()-1);
				
				Rational x_int_tmp = curve1.segments[i].getXIntersectionWith( curve2.segments[j] );
				
				if( x_int_tmp.equals( Rational.getNaN() ) ) {
					break;
				}
				
				// TODO I think you can already stop as soon as you found the first intersection
				// 		or you could return a list of intersections after the whole search
				
				// TODO do I need to take && or &?
				if( x_int_tmp.greater( new Rational( 0.0 ) ) ){
					// test if the curves is defined by these segments at x_tmp so it really is a valid intersection 
					if( !curve1_last ){
						if( !curve2_last ){
							if( x_int_tmp.less( curve1.segments[i+1].x0 )
								&& x_int_tmp.less( curve1.segments[j+1].x0 )
								&& x_int_tmp.less( x_int ) ) {				
										
									x_int = x_int_tmp;
							}
						} else {
							if( x_int_tmp.less( curve1.segments[i+1].x0 )
								&& x_int_tmp.less( x_int ) ) {				
										
									x_int = x_int_tmp;
							}
						}
					} else {
						if( !curve2_last ){
							if( x_int_tmp.less( curve1.segments[j+1].x0 )
								&& x_int_tmp.less( x_int ) ) {				
										
									x_int = x_int_tmp;
							}
						} else {
							if( x_int_tmp.less( x_int )) {				
										
									x_int = x_int_tmp;
							}
						}
					}
				}
			}
		}
		return x_int;
	}

	/**
	 * Returns a string representation of this curve.
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		String result = "{";
		for (int i = 0; i < segments.length; i++) {
			if (i > 0) {
				result += ",";
			}
			if (segments[i].leftopen) {
				result += "!";
			}
			result += "(" + segments[i].x0 + "," + segments[i].y0 + "),"
						+ segments[i].grad;
		}
		result += "}";
		return result;
	}

	/**
	 * Prints the string representation of this curve to stdout.
	 */
	public void dump() {
		System.out.print(this);
	}
	
	public int getMaximumItemCount() { return segments.length; }
	
	public Rational getX(int index) { return segments[index].x0; }
	
	public Rational getY(int index) { return segments[index].y0; }

	public int getItemCount() { return segments.length; }

	public List<LinearSegment> getItems() {
		return Collections.unmodifiableList(Arrays.asList(segments));
	}

	public int indexOf( Rational x ) {
		int s = getSegmentDefining( x );
		if( segments[s].x0 == x ) {
			return s;
		} else {
			return -1;
		}
	}
	
	/**
	 * Common helper for computing a new curve.
	 * @param c1
	 * @param c2
	 * @param operator
	 * @return
	 */
	private static Curve computeResultingCurve( Curve c1, Curve c2, int operator ) {
		switch( operator ) {
		case OPERATOR_ADD:
			if( c1.equals( ZERO_DELAY_BURST_CURVE ) || c2.equals( ZERO_DELAY_BURST_CURVE ) ){
				return ZERO_DELAY_BURST_CURVE.copy();
			}
			break;
		case OPERATOR_SUB:
			if( c1.equals( ZERO_DELAY_BURST_CURVE ) || c2.equals( ZERO_DELAY_BURST_CURVE ) ){
				return ZERO_DELAY_BURST_CURVE.copy();
			}
			break;
		case OPERATOR_MIN:
			if( c1.equals( ZERO_DELAY_BURST_CURVE ) ){
				return c2.copy();
			}
			if( c2.equals( ZERO_DELAY_BURST_CURVE ) ){
				return c1.copy();
			}
			break;
		case OPERATOR_MAX:
			if( c1.equals( ZERO_DELAY_BURST_CURVE ) || c2.equals( ZERO_DELAY_BURST_CURVE ) ){
				return ZERO_DELAY_BURST_CURVE.copy();
			}
			break;
		default:
		}
		
		Curve result = new Curve();

		Rational x = new Rational( 0.0 );

		Rational x_cross;

		boolean leftopen;

		int i1 = 0;
		int i2 = 0;
		while( i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount() ) {
			Rational x_next1 = ( i1+1 < c1.getSegmentCount() ) ?
								c1.getSegment( i1+1 ).x0 : Rational.getPosInfinity();
			Rational x_next2 = ( i2+1 < c2.getSegmentCount() ) ?
								c2.getSegment( i2+1 ).x0 : Rational.getPosInfinity();
			Rational x_next  = Rational.min( x_next1, x_next2 );
								
			leftopen = c1.getSegment( i1 ).leftopen || c2.getSegment( i2 ).leftopen;

			switch( operator ) {
			case OPERATOR_ADD:
				result.addSegment( add( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen ) );
				break;
			case OPERATOR_SUB:
				result.addSegment( sub( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen ) );
				break;
			case OPERATOR_MIN:
				x_cross = c1.getSegment( i1 ).getXIntersectionWith( c2.getSegment( i2 ) );
				if ( x.less( x_cross ) && x_cross.less( x_next ) ) {
					result.addSegment( min( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen, false ) );
					result.addSegment( min( c1.getSegment( i1 ), c2.getSegment( i2 ), x_cross, false, true ) );
				} else {
					result.addSegment( min( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen, false ) );
				}
				break;
			case OPERATOR_MAX:
				x_cross = c1.getSegment( i1 ).getXIntersectionWith( c2.getSegment( i2 ) );
				if ( x.less( x_cross ) && x_cross.less( x_next ) ) {
					result.addSegment( max( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen, false ) );
					result.addSegment( max( c1.getSegment( i1 ), c2.getSegment( i2 ), x_cross, false, true ) );
				} else {
					result.addSegment( max( c1.getSegment( i1 ), c2.getSegment( i2 ), x, leftopen, false ) );
				}
				break;
			default:
			}

			if ( x_next1.equals( x_next ) ) {
				i1++;
			}
			if ( x_next2.equals( x_next ) ) {
				i2++;
			}
			x = x_next;
		}

		result.beautify();

		if ( Configuration.createLabels() ) {
			String label1 = c1.getAlias();
			String label2 = c2.getAlias();
			switch( operator ) {
			case OPERATOR_ADD:
				result.setLabel( "(" + ( ( label1.length() > 0 ) ? label1 : "?" )
				             + "+" + ( ( label2.length() > 0 ) ? label2 : "?" ) + ")" );
				break;
			case OPERATOR_SUB:
				result.setLabel( "(" + ( ( label1.length() > 0 ) ? label1 : "?" )
				             + "-" + ( ( label2.length() > 0) ? label2 : "?" ) + ")" );
				break;
			case OPERATOR_MIN:
				result.setLabel( "(" + ( ( label1.length() > 0 ) ? label1 : "?" )
				             + "\u2227" + ( ( label2.length() > 0) ? label2 : "?" ) + ")" );
				break;
			case OPERATOR_MAX:
				result.setLabel( "(" + ( ( label1.length() > 0 ) ? label1 : "?" )
				             + "\u2228" + ( ( label1.length() > 0) ? label1 : "?" ) + ")" );
				break;
			default:
			}
		}

		return result;
	}
	
	/**
	 * Helper creating a new segment starting at x0 that is the sum of the given getSegment.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private static LinearSegment add( LinearSegment s1, LinearSegment s2, Rational x, boolean leftopen ) {
		LinearSegment result = LinearSegment.createHorizontalLine( 0.0 );
		result.x0       = x;
		result.y0       = Rational.add( s1.f( x ), s2.f( x ) );
		result.grad     = Rational.add( s1.grad, s2.grad );
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the difference between the given getSegment.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private static LinearSegment sub( LinearSegment s1, LinearSegment s2, Rational x, boolean leftopen ) {
		LinearSegment result = LinearSegment.createHorizontalLine( 0.0 );
		result.x0       = x;
		result.y0       = Rational.sub( s1.f( x ), s2.f( x ) );
		result.grad     = Rational.sub( s1.grad, s2.grad );
		result.leftopen = leftopen;
		return result;
	}

	/**
	 * Helper creating a new segment starting at x0 that is the minimum of the given getSegment.
	 * Note: Only valid if s1 and s2 do not intersect before the next IP.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private static LinearSegment min( LinearSegment s1, LinearSegment s2, Rational x, boolean leftopen, boolean crossed ) {
		Rational f1_x = s1.f( x );
		Rational f2_x = s2.f( x );

		LinearSegment result = LinearSegment.createHorizontalLine( 0.0 );
		result.x0 = x;
		if ( crossed || Rational.abs( Rational.sub( f1_x, f2_x ) ).less( EPSILON ) ) {
			result.y0   = f1_x;
			result.grad = Rational.min( s1.grad, s2.grad );
		} else if ( f1_x.less( f2_x ) ) {
			result.y0   = f1_x;
			result.grad = s1.grad;
		} else {
			result.y0   = f2_x;
			result.grad = s2.grad;
		}
		result.leftopen = leftopen;
		return result;
	}
	
	/**
	 * Returns a curve that is the sum of this curve and the given curve.
	 * @param curve2 the given curve.
	 * @return the sum of curves
	 */
	public static Curve add( Curve curve1, Curve curve2 ) {
		return computeResultingCurve( curve1, curve2, OPERATOR_ADD );
	}

	/**
	 * Returns a curve that is the difference between this curve and the given curve.
	 * @param curve2 the given curve.
	 * @return the difference between curves
	 */
	public static Curve sub( Curve curve1, Curve curve2 ) {
		return computeResultingCurve( curve1, curve2, OPERATOR_SUB );
	}

	/**
	 * Returns a curve that is the minimum of this curve and the given curve.
	 * @param curve2 the given curve.
	 * @return the minimum of curves
	 */
	public static Curve min( Curve curve1, Curve curve2 ) {
		return computeResultingCurve( curve1, curve2, OPERATOR_MIN );
	}

	/**
	 * Returns a curve that is the maximum of this curve and the given curve.
	 * @param curve2 the given curve.
	 * @return the maximum of curves
	 */
	public static Curve max( Curve curve1, Curve curve2 ) {
		return computeResultingCurve( curve1, curve2, OPERATOR_MAX );
	}

	/**
	 * Helper creating a new segment starting at x0 that is the maximum of the given segments.
	 * Note: Only valid if s1 and s2 do not intersect before the next IP.
	 * @param s1
	 * @param s2
	 * @param x
	 * @param leftopen
	 * @return
	 */
	private static LinearSegment max( LinearSegment s1, LinearSegment s2, Rational x, boolean leftopen, boolean crossed ) {
		Rational f1_x = s1.f( x );
		Rational f2_x = s2.f( x );

		LinearSegment result = LinearSegment.createHorizontalLine( 0.0 );
		result.x0 = x;
		if ( crossed || Rational.abs( Rational.sub( f1_x, f2_x ) ).less( EPSILON ) ) {
			result.y0   = f1_x;
			result.grad = Rational.max( s1.grad, s2.grad );
		} else if ( f1_x.greater( f2_x ) ) {
			result.y0   = f1_x;
			result.grad = s1.grad;
		} else {
			result.y0   = f2_x;
			result.grad = s2.grad;
		}
		result.leftopen = leftopen;
		return result;
	}
	
	/**
	 * Returns a copy of curve bounded at the x-axis.
	 * @return the bounded curve.
	 */
	public static Curve boundAtXAxis( Curve curve ) {
		Curve result = new Curve();

		LinearSegment s;
		for ( int i = 0; i < curve.getSegmentCount(); i++ ) {
			if ( curve.getSegment( i ).y0.greater( new Rational( 0.0 ) ) ) {
				result.addSegment( curve.getSegment( i ).copy() );

				if ( curve.getSegment(i).grad.less( new Rational( 0.0 ) ) ) {
					Rational x_cross = curve.getSegment( i ).getXIntersectionWith( LinearSegment.X_AXIS );
					if ( i+1 >= curve.getSegmentCount() || x_cross.less( curve.getSegment( i+1 ).x0 ) ) {
						s = LinearSegment.createHorizontalLine( 0.0 );
						s.x0 = x_cross;
						result.addSegment( s );
					}
				}
			} else {
				s = LinearSegment.createHorizontalLine( 0.0 );
				s.x0       = curve.getSegment( i ).x0;
				s.leftopen = curve.getSegment( i ).leftopen;
				result.addSegment( s );

				if ( curve.getSegment(i).grad.greater( new Rational( 0.0 ) ) ) {
					Rational x_cross = curve.getSegment( i ).getXIntersectionWith( LinearSegment.X_AXIS );
					if ( i+1 >= curve.getSegmentCount() || x_cross.less( curve.getSegment(i+1).x0 ) ) {
						s = LinearSegment.createHorizontalLine( 0.0 );
						s.x0   = x_cross;
						s.grad = curve.getSegment( i ).grad;
						result.addSegment( s );
					}
				}
			}
		}
		result.beautify();
		if ( Configuration.createLabels() ) {
			result.setLabel( "[" + curve.getAlias() + "]\u207A" );
		}
		return result;
	}	

	/**
	 * Returns the vertical deviation between the given two curves.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return the value of the vertical deviation.
	 */
	public static Rational getVerticalDeviation( Curve c1, Curve c2 ) {
		if ( c1.getSustainedRate().greater( c2.getSustainedRate() ) ) {
			return Rational.getPosInfinity();
		}

		Rational result = Rational.getNegInfinity();
		ArrayList<Rational> xcoords = computeInflectionPointsX( c1, c2 );		
		for( int i = 0; i < xcoords.size(); i++ ) {
			Rational ip_x = ( (Rational) xcoords.get( i ) );

			Rational backlog = Rational.sub( c1.f( ip_x ), c2.f( ip_x ) );
			result = Rational.max( result, backlog );
		}
		return result;
	}

	/**
	 * Returns the horizontal deviation between the given two curves.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return the value of the horizontal deviation.
	 */
	public static Rational getHorizontalDeviation( Curve c1, Curve c2 ) {
		if ( c1.getSustainedRate().greater( c2.getSustainedRate() ) ) {
			return Rational.getPosInfinity();
		}

		Rational result = Rational.getNegInfinity();
		for( int i = 0; i < c1.getSegmentCount(); i++ ) {
			Rational ip_y = c1.getSegment( i ).y0;

			Rational delay = Rational.sub( c2.f_inv( ip_y, true ), c1.f_inv( ip_y, false ) );
			result = Rational.max( result, delay );
		}
		for( int i = 0; i < c2.getSegmentCount(); i++ ) {
			Rational ip_y = c2.getSegment( i ).y0;

			Rational delay = Rational.sub( c2.f_inv( ip_y, true ), c1.f_inv( ip_y, false ) );
			result = Rational.max( result, delay );
		}
		return result;
	}

	/**
	 * Returns an <code>ArrayList</code> instance of those x-coordinates
	 * at which either c1 or c2 or both have an inflection point. There
	 * will be multiple occurences of an x-coordinate, if at least one
	 * curve has a discontinuity at that x-coordinate.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return an <code>ArrayList</code> of <code>Double</code> objects
	 * containing the x-coordinates of the respective inflection point.
	 */
	public static ArrayList<Rational> computeInflectionPointsX( Curve c1, Curve c2 ) {
		ArrayList<Rational> xcoords = new ArrayList<Rational>();

		int i1 = 0;
		int i2 = 0;
		while( i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount() ) {
			Rational x1 = ( i1 < c1.getSegmentCount() ) ?
						c1.getSegment( i1 ).x0 : Rational.getPosInfinity();
			Rational x2 = ( i2 < c2.getSegmentCount() ) ?
						c2.getSegment( i2 ).x0 : Rational.getPosInfinity();
			if ( x1.less( x2 ) ) {
				xcoords.add( x1.copy() );
				i1++;
			} else if ( x1.greater( x2 ) ) {
				xcoords.add( x2.copy() );
				i2++;
			} else {
				xcoords.add( x1.copy() );
				i1++;
				i2++;
			}
		}
		return xcoords;
	}

	/**
	 * Returns an <code>ArrayList</code> instance of those y-coordinates
	 * at which either c1 or c2 or both have an inflection point.
	 * @param c1 the first curve.
	 * @param c2 the second curve.
	 * @return an <code>ArrayList</code> of <code>Double</code> objects
	 * containing the x-coordinates of the respective inflection point.
	 */
	public static ArrayList<Rational> computeInflectionPointsY( Curve c1, Curve c2 ) {
		ArrayList<Rational> ycoords = new ArrayList<Rational>();

		int i1 = 0;
		int i2 = 0;
		while( i1 < c1.getSegmentCount() || i2 < c2.getSegmentCount() ) {
			Rational y1 = ( i1 < c1.getSegmentCount() ) ?
						c1.getSegment( i1 ).y0 : Rational.getPosInfinity();
			Rational y2 = ( i2 < c2.getSegmentCount() ) ?
						c2.getSegment( i2 ).y0 : Rational.getPosInfinity();
			if ( y1.less( y2 ) ) {
				ycoords.add( y1.copy() );
				i1++;
			} else if ( y1.greater( y2 ) ) {
				ycoords.add( y2.copy() );
				i2++;
			} else {
				ycoords.add( y1.copy() );
				i1++;
				i2++;
			}
		}

		return ycoords;
	}
	
	public boolean equals( Curve other ) {
		Curve this_cpy = this.copy();
		Curve other_cpy = other.copy();
		
		this_cpy.beautify();
		other_cpy.beautify();
		
		int this_segment_length = this_cpy.segments.length;
		
		if( this_segment_length != other_cpy.segments.length ){
			return false;
		}
		
		for( int i = 0; i < this_segment_length; i++ ) {
			if( !this_cpy.segments[i].equals( other_cpy.segments[i] ) ){
				return false;
			}
		}
			
		return true;
	}
}