/*
 *  (c) 2008 Andreas Kiefer, disco | Distributed Computer Systems Lab
 *                                   University of Kaiserslautern, Germany
 *      All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import unikl.disco.misc.Rational;
import unikl.disco.misc.SetUtils;

import unikl.disco.curves.Curve;

import unikl.disco.nc.Configuration.Flag;

import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Vertex;

import org.apache.commons.math3.util.Pair;

/**
 * This class contains all members and methods that are needed for more than one
 * analysis. Those were formerly contained in <code>NetworkAnalyser</code>
 * but are extracted to provide better usability and maintenance.
 * 
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 */
public abstract class Analysis
{
	protected Network			network							= null;
	
	protected Flow				flow_of_interest				= null;

	public abstract boolean performAnalysis( Flow flow_of_interest );

	/**
	 * Returns the backlog bound of the given two curves.
	 * @param alpha the arrival curve.
	 * @param beta  the service curve.
	 * @return the value of the backlog bound.
	 */
	public abstract Rational getBacklogBound();
	
	/**
	 * Returns the delay bound of the given two curves.
	 * @param alpha the arrival curve.
	 * @param beta  the service curve.
	 * @return the value of the delay bound.
	 */
	public abstract Rational getDelayBound();

	/**
	 * Returns the output bound for a set of <code>outgoing_flows</code> from
	 * the given <code>server</code>.<br/>
	 * It is possible to change between iterative and recursive computation by
	 * calling <code>setUseRecursiveOutputBoundComputation()</code>.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param outgoing_flows
	 *            set of flows for which the output bound is computed
	 * @return a <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 * @author Andreas Kiefer
	 */
	protected Curve computeOutputBound( Vertex server, Set<Flow> outgoing_flows ) throws OutOfMemoryError
	{
		// TODO make impact of FIFO flag triggering compute FIFO service curve on TFA, SFA as well as PMOO clear!
		if ( Configuration.useRecursiveOutputBoundComputation() )
		{
			return computeRecursiveOutputBound( server, outgoing_flows );
		}
		else
		{
			return computeIterativeOutputBound( server, outgoing_flows );
		}
	}

	/**
	 * Returns the output bound for a set of <code>outgoing_flows</code> from
	 * the given <code>server</code> by recursion.<br/>
	 * It does so by first iterating over all incoming edges of the server,
	 * summing up the output bounds for all flows of interest (incoming flows
	 * contained also in <code>outgoing_flows</code>) and all cross flows
	 * (incoming flows not contained in <code>outgoing_flows</code>).<br/>
	 * In a next step, the cross flows are factored out of the service curve to
	 * obtain the effective service curve. Per default, blind multiplexing is
	 * assumed. To use FIFO service instead you need to invoke
	 * <code>setUseFIFOService(true)</code> on the analyser instance before
	 * computing any bounds.<br/>
	 * The output bound is then computed using arrival and service curves only.
	 * Invoke <code>setUseGamma(true)</code> on the analyser instance to
	 * globally activate the additional use of the maximum service curve for
	 * output bound computation. Finally, by invoking
	 * <code>setUseExtraGamma(true)</code> it is possible to convolve the
	 * obtained output bound once more with the maximum service curve whose
	 * latency has been removed before returning the result.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param outgoing_flows
	 *            set of flows for which the output bound is computed
	 * @return a <code>Curve</code> instance containing the output bound
	 * @see NetworkAnalyser#setUseFifoService(boolean)
	 * @see NetworkAnalyser#setUseGamma(boolean)
	 * @see NetworkAnalyser#setUseExtraGamma(boolean)
	 * @throws OutOfMemoryError
	 */
	protected Curve computeRecursiveOutputBound( Vertex server, Set<Flow> outgoing_flows ) throws OutOfMemoryError
	{
		Curve alpha_out = Curve.createNullCurve();

		if ( outgoing_flows.size() == 0 )
		{
			return alpha_out;
		}

		Set<Flow> incoming_flows = network.getSourceFlowSet( server );
		Set<Flow> included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
		Set<Flow> excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
		excluded_flows.remove( flow_of_interest );

		Curve sum_alpha = network.getAggregateSourceFlowAC( server, included_flows );
		Curve sum_alpha_excl = network.getAggregateSourceFlowAC( server, excluded_flows );

		for ( Vertex predecessor : network.getPredecessors( server ) )
		{
			incoming_flows = network.getIncomingFlows( predecessor, server );
			included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
			excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
			excluded_flows.remove( flow_of_interest );

			sum_alpha = Curve.add( sum_alpha, computeRecursiveOutputBound( predecessor, included_flows ) );
			sum_alpha_excl = Curve.add( sum_alpha_excl, computeRecursiveOutputBound( predecessor, excluded_flows ) );
		}

		Curve beta_effective = server.getServiceCurve();
		Curve gamma = server.getMaxServiceCurve();
		if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_ON ||
				( Configuration.getUseArbMuxFlag() == Flag.NODELOCAL_FLAG & server.useArbitraryMultiplexing() ) )
		{
			beta_effective = Curve.boundAtXAxis( Curve.sub( beta_effective, sum_alpha_excl ) );
		}
		else
		{
			beta_effective = computeFifoServiceCurve( sum_alpha_excl, beta_effective );
		}

		if( Configuration.getUseGammaFlag() == Flag.GLOBALLY_ON ||
			( Configuration.getUseGammaFlag() == Flag.NODELOCAL_FLAG & server.useGamma() ) )
		{
			alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( sum_alpha, gamma ), beta_effective ) );
		}
		else
		{
			alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( sum_alpha, beta_effective ) );
		}
		
		if( Configuration.getUseExtraGammaFlag() == Flag.GLOBALLY_ON ||
				( Configuration.getUseExtraGammaFlag() == Flag.NODELOCAL_FLAG & server.useExtraGamma() ) )
		{
			alpha_out = MinPlus.convolve( alpha_out, gamma.removeLatency() );
		}
		
		return alpha_out;
	}

	/**
	 * Returns the output bound for a set of <code>outgoing_flows</code> from
	 * the given <code>server</code> by iteration.<br/>
	 * For the iteration the help of a stack is required, to track the tuples of
	 * <code>server</code> and <code>outgoing_flows</code>. Also a table is
	 * needed for mapping the tuple to the result.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param outgoing_flows
	 *            set of flows for which the output bound is computed
	 * @return a <code>Curve</code> instance containing the output bound
	 * @see NetworkAnalyser#decrementOutputBoundStack(Stack, Map)
	 * @see NetworkAnalyser#incrementOutputBoundStack(Stack, Map)
	 * @throws OutOfMemoryError
	 * @author Andreas Kiefer
	 */
	protected Curve computeIterativeOutputBound( Vertex server, Set<Flow> outgoing_flows ) throws OutOfMemoryError
	{
		Stack<Pair<Vertex, Set<Flow>>> stack = new Stack<Pair<Vertex, Set<Flow>>>();
		Map<Pair<Vertex, Set<Flow>>, Curve> results = new HashMap<Pair<Vertex, Set<Flow>>, Curve>();

		Pair<Vertex, Set<Flow>> next = new Pair<Vertex, Set<Flow>>( server, outgoing_flows );
		stack.push( next );
		// until everything is computed
		while ( !stack.isEmpty() )
		{
			if ( !decrementOutputBoundStack( stack, results ) )
			{
				incrementOutputBoundStack( stack, results );
			}
		}
		Curve alpha_out = results.get( next );
		if ( alpha_out == null )
		{
			alpha_out = Curve.createNullCurve();
		}

		return alpha_out;
	}

	/**
	 * Puts all required output bound calculations for a given
	 * <code>server</code> on a <code>stack</code><br/>
	 * When the output bound for a given <code>server</code> and a set of
	 * <code>outgoing_flows</code> cannot be computed, it's because the
	 * calculation depended on results that are not yet computed, so all
	 * dependencies have to be put on the stack. That means, iterating over all
	 * the incoming edges of the server.
	 * 
	 * @param stack
	 *            the stack with the <code>server</code>,
	 *            <code>outgoing_flows</code>-tuple.
	 * @param results
	 *            maps the tuple to the result.
	 * @see NetworkAnalyser#computeRecursiveOutputBound(boolean)
	 * @see NetworkAnalyser#computeIterativeOutputBound(Vertex, Set)
	 * @see NetworkAnalyser#decrementOutputBoundStack(Stack, Map)
	 * @see NetworkAnalyser#isAtEnd(Vertex, Set)
	 * @author Andreas Kiefer
	 */
	protected void incrementOutputBoundStack(	Stack<Pair<Vertex, Set<Flow>>> stack,
												Map<Pair<Vertex, Set<Flow>>, Curve> results )
	{
		Pair<Vertex, Set<Flow>> next = stack.peek();

		Vertex server = (Vertex) next.getFirst();
		Set<Flow> outgoing_flows = next.getSecond();
		
		Set<Flow> incoming_flows = network.getSourceFlowSet( server );
		Set<Flow> included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
		Set<Flow> excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
		excluded_flows.remove( flow_of_interest );

		for ( Vertex predecessor : network.getPredecessors( server ) )
		{
			incoming_flows = network.getIncomingFlows( predecessor, server );
			included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
			excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
			excluded_flows.remove( flow_of_interest );

			// is the required output bound already computed?
			Pair<Vertex, Set<Flow>> p_incl = new Pair<Vertex, Set<Flow>>( predecessor, included_flows );
			if ( !results.containsKey( p_incl ) )
			{
				stack.push( p_incl );
				// some output bounds doesn't depend on anything
				if ( isAtEnd( predecessor, included_flows ) )
				{
					decrementOutputBoundStack( stack, results );
				}
			}
			// is the required output bound already computed?
			Pair<Vertex, Set<Flow>> p_excl = new Pair<Vertex, Set<Flow>>( predecessor, excluded_flows );
			if ( !results.containsKey( p_excl ) )
			{
				stack.push( p_excl );
				// some output bounds doesn't depend on anything
				if ( isAtEnd( predecessor, excluded_flows ) )
				{
					decrementOutputBoundStack( stack, results );
				}
			}
		}
	}

	/**
	 * Recognises if a <code>server</code> needn't be processed any further.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param outgoing_flows
	 *            set of flows for which the output bound is computed
	 * @return if the computation is at the end.
	 * @see NetworkAnalyser#incrementOutputBoundStack(Stack, Map)
	 * @author Andreas Kiefer
	 */
	protected boolean isAtEnd( Vertex server, Set<Flow> outgoing_flows )
	{
		if ( outgoing_flows.size() == 0 )
		{
			return true;
		}
		if ( network.inDegree(server) == 0 )
		{
			return true;
		}
		return false;
	}

	/**
	 * Returns the output bound for a set of <code>outgoing_flows</code> from
	 * the given <code>server</code> by recursion.<br/>
	 * This computation is nearly the same as the recursive in
	 * <code>computeRecursiveOutputBound(boolean)</code>. But the output bound
	 * is computed for the element on top of a stack. This element is a tuple
	 * with the server and the flows of interests passed as the arguments in the
	 * recursion. In many cases, the computation depends on previously computed
	 * output bounds, so when these are lacking, the calling method has to be
	 * notified. After a successful computation, the result is added to a map.
	 * 
	 * @param stack
	 *            the stack with the <code>server</code>,
	 *            <code>outgoing_flows</code>-tuple.
	 * @param results
	 *            maps the tuple to the result.
	 * @return if the output bound of the element on top of the stack could be
	 *         computed.
	 * @see NetworkAnalyser#computeRecursiveOutputBound(boolean)
	 * @see NetworkAnalyser#computeIterativeOutputBound(Vertex, Set)
	 * @see NetworkAnalyser#incrementOutputBoundStack(Stack, Map)
	 * @author Andreas Kiefer
	 */
	protected boolean decrementOutputBoundStack(	Stack<Pair<Vertex, Set<Flow>>> stack,
													Map<Pair<Vertex, Set<Flow>>, Curve> results )
	{
		Pair<Vertex, Set<Flow>> next = stack.pop();

		Vertex server = (Vertex) next.getFirst();
		Set<Flow> outgoing_flows = (Set<Flow>) next.getSecond();

		Curve alpha_out = Curve.createNullCurve();

		if ( outgoing_flows.size() == 0 )
		{
			results.put( next, alpha_out );
			return true;
		}
			
		Set<Flow> incoming_flows = network.getSourceFlowSet( server );
		Set<Flow> included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
		Set<Flow> excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
		excluded_flows.remove( flow_of_interest );

		Curve sum_alpha = network.getAggregateSourceFlowAC( server, included_flows );
		Curve sum_alpha_excl = network.getAggregateSourceFlowAC( server, excluded_flows );
		
		for ( Vertex predecessor : network.getPredecessors(server) )
		{
			incoming_flows = network.getIncomingFlows( predecessor, server );
			included_flows = SetUtils.getIntersection( incoming_flows, outgoing_flows );
			excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
			excluded_flows.remove( flow_of_interest );

			Curve alpha_incl = results.get( new Pair<Vertex, Set<Flow>>( predecessor, included_flows ) );
			if ( alpha_incl == null )
			{
				stack.push( next );
				return false;
			}
			Curve alpha_excl = results.get( new Pair<Vertex, Set<Flow>>( predecessor, excluded_flows ) );
			if ( alpha_excl == null )
			{
				stack.push( next );
				return false;
			}

			sum_alpha = Curve.add( sum_alpha, alpha_incl );
			sum_alpha_excl = Curve.add( sum_alpha_excl, alpha_excl );
		}

		Curve beta_effective = server.getServiceCurve();
		Curve gamma = server.getMaxServiceCurve();
		if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_ON ||
				( Configuration.getUseArbMuxFlag() == Flag.NODELOCAL_FLAG & server.useArbitraryMultiplexing() ) )
		{
			beta_effective = Curve.boundAtXAxis( Curve.sub( beta_effective, sum_alpha_excl ) );
		}
		else
		{
			beta_effective = computeFifoServiceCurve( sum_alpha_excl, beta_effective );
		}

		if( Configuration.getUseGammaFlag() == Flag.GLOBALLY_ON ||
				( Configuration.getUseGammaFlag() == Flag.NODELOCAL_FLAG & server.useGamma() ) )
		{
			alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( MinPlus.convolve( sum_alpha, gamma ), beta_effective) );
		}
		else
		{
			alpha_out = Curve.makeArrivalCurve( MinPlus.deconvolve( sum_alpha, beta_effective ) );
		}
		
		if( Configuration.getUseExtraGammaFlag() == Flag.GLOBALLY_ON ||
				( Configuration.getUseExtraGammaFlag() == Flag.NODELOCAL_FLAG & server.useExtraGamma() ) )
		{
			alpha_out = MinPlus.convolve( alpha_out, gamma.removeLatency() );
		}

		results.put( next, alpha_out );
		return true;
	}
	
	/**
	 * Computes the effective FIFO service curve for a server with the service curve
	 * <code>beta</code> experiencing cross-traffic with arrival curve <code>alpha</code>.
	 * @param alpha arrival curve of cross-traffic
	 * @param beta server's service curve
	 * @return the FIFO service curve
	 */
	public static Curve computeFifoServiceCurve( Curve alpha, Curve beta ) {
		if ( Configuration.performChecks() && !alpha.isConcave() ) {
			throw new IllegalArgumentException( "First curve must be concave!" );
		}
		if ( Configuration.performChecks() && !beta.isConvex() ) {
			throw new IllegalArgumentException( "Second curve must be convex!" );
		}

		Rational last_y = alpha.fLimitRight( new Rational( 0.0 ) );

		ArrayList<Rational> ycoords = Curve.computeInflectionPointsY( alpha, beta );
		for( int i = 0; i < ycoords.size(); i++ ) {
			Rational ip_y = ( ycoords.get( i ) );
			if ( ip_y.less( last_y ) ) {
				continue;
			}

			Rational x_alpha = alpha.f_inv( ip_y, false );
			Rational x_beta  = beta.f_inv( ip_y, true );
			if ( alpha.getGradientLimitRight( x_alpha ).le( beta.getGradientLimitRight( x_beta ) ) ) {
				
				Rational theta = Rational.sub( x_beta, x_alpha );
				Curve beta_fifo = Curve.boundAtXAxis( 
										Curve.min(	Curve.sub( beta,alpha.shiftRight( theta ) ), 
	    											Curve.createBurstDelay( x_beta ) ) );
	    		return beta_fifo;
			}
		}
		return beta.copy();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return getClass().getSimpleName();
	}
}
