package unikl.disco.nc;

import unikl.disco.curves.Curve;
import unikl.disco.curves.LinearSegment;
import unikl.disco.misc.Rational;

public class MinPlus {
	// ---------------------------------------------------------------------
	// convolution
	// ---------------------------------------------------------------------	
	public static Curve convolve( Curve curve1, Curve curve2 ) throws IllegalArgumentException {
		if ( curve1.isAlmostConcave() || curve2.isAlmostConcave() ) {
			return convolveAlmostConcave( curve1, curve2 );
		}
		
		if ( curve1.isConvex() || curve2.isConvex() ) {
			return convolveConvex( curve1, curve2 );
		}
		
		throw new IllegalArgumentException("Both curves to be convolved must either be almost concave or convex!");
	}
	
	/**
	 * Returns the convolution of two curves, which must be (almost) concave
	 * @param curve1 the first curve to convolve with.
	 * @param curve2 the second curve to convolve with.
	 * @return the convolved curve.
	 */
	private static Curve convolveAlmostConcave( Curve curve1, Curve curve2 ) {
		Rational L1 = curve1.getLatency();
		Rational L2 = curve2.getLatency();
		if ( L1.equals( Rational.getPosInfinity() ) ) {
			return curve2.copy();
		}
		if ( L2.equals( Rational.getPosInfinity() ) ) {
			return curve1.copy();
		}

		Curve result = ( Curve.min( curve1.removeLatency(), curve2.removeLatency() ) ).shiftRight( Rational.add( L1, L2 ) );
		result.beautify();
		if ( Configuration.createLabels()
				&& (curve1.getLabel() != null || curve1.getAlias() != null
						|| curve2.getLabel() != null || curve2.getAlias() != null)) {
			String label1 = (curve1.getAlias().length() > 0) ? curve1.getAlias() : "?";
			String label2 = (curve2.getAlias().length() > 0) ? curve2.getAlias() : "?";
			result.setLabel( "(" + label1 + "\u2297" + label2 + ")" );
		}
		return result;
	}

	/**
	 * Returns the convolution of two curve, which must be convex
	 * @param curve1 first the curve to convolve with.
	 * @param curve2 second the curve to convolve with.
	 * @return the convolved curve.
	 */
	private static Curve convolveConvex( Curve curve1, Curve curve2 ) {
		if ( curve1.isBurstDelay() && curve1.getSegment(1).x0.equals( new Rational( 0.0 ) ) ) {
			return curve2.copy();
		}
		if ( curve2.isBurstDelay() && curve2.getSegment(1).x0.equals( new Rational( 0.0 ) ) ) {
			return curve1.copy();
		}

		Curve result = new Curve();

		LinearSegment s;
		s = LinearSegment.createHorizontalLine( 0.0 );
		s.x0 = new Rational( 0.0 );
		s.y0 = Rational.add( curve1.f( new Rational( 0.0 ) ), curve2.f( new Rational( 0.0 ) ) );
		result.addSegment(s);

		int i1 = (curve1.isRealDiscontinuity(0)) ? 1 : 0;
		int i2 = (curve2.isRealDiscontinuity(0)) ? 1 : 0;
		if (i1 > 0 || i2 > 0) {
			s = LinearSegment.createHorizontalLine( 0.0 );
			s.x0       = new Rational( 0.0 );
			s.y0       = Rational.add( curve1.fLimitRight( new Rational( 0.0 ) ), curve2.fLimitRight( new Rational( 0.0 ) ) );
			s.leftopen = true;
			result.addSegment(s);
		}

		while(i1 < curve1.getSegmentCount() || i2 < curve2.getSegmentCount()) {
			if ( curve1.getSegment(i1).grad.less( curve2.getSegment(i2).grad ) ) {
				if (i1+1 >= curve1.getSegmentCount()) {
					result.getSegment(result.getSegmentCount()-1).grad = curve1.getSegment(i1).grad;
					break;
				}

				s = LinearSegment.createHorizontalLine( 0.0 );
				s.x0 = Rational.add( result.getSegment( result.getSegmentCount()-1 ).x0,
						( Rational.sub( curve1.getSegment( i1+1 ).x0, curve1.getSegment( i1 ).x0 ) ) );
				s.y0 = Rational.add( result.getSegment( result.getSegmentCount()-1 ).y0,
						( Rational.sub( curve1.getSegment( i1+1 ).y0, curve1.getSegment( i1 ).y0 ) ) );
				result.getSegment(result.getSegmentCount()-1).grad = curve1.getSegment(i1).grad;
				result.addSegment(s);
				i1++;
			} else {
				if (i2+1 >= curve2.getSegmentCount()) {
					result.getSegment(result.getSegmentCount()-1).grad = curve2.getSegment(i2).grad;
					break;
				}

				s = LinearSegment.createHorizontalLine( 0.0 );
				s.x0 = Rational.add( result.getSegment( result.getSegmentCount()-1 ).x0,
						( Rational.sub( curve2.getSegment( i2+1 ).x0, curve2.getSegment( i2 ).x0 ) ) );
				s.y0 = Rational.add( result.getSegment( result.getSegmentCount()-1 ).y0,
						( Rational.sub( curve2.getSegment( i2+1 ).y0, curve2.getSegment( i2 ).y0 ) ) );
				result.getSegment(result.getSegmentCount()-1).grad = curve2.getSegment(i2).grad;
				result.addSegment(s);
				i2++;
			}
		}
		result.beautify();
		if ( Configuration.createLabels()
				&& (curve1.getLabel() != null || curve1.getAlias() != null
						|| curve2.getLabel() != null || curve2.getAlias() != null)) {
			String label1 = (curve1.getAlias().length() > 0) ? curve1.getAlias() : "?";
			String label2 = (curve2.getAlias().length() > 0) ? curve2.getAlias() : "?";
			result.setLabel ( "(" + label1 + "\u2297" + label2 + ")" );
		}
		return result;
	}

	// ---------------------------------------------------------------------
	// deconvolution
	// ---------------------------------------------------------------------
	public static Curve deconvolve( Curve curve1, Curve curve2 ) throws IllegalArgumentException {
		if ( Configuration.performChecks() && !curve1.isAlmostConcave() ) {
			throw new IllegalArgumentException("First curve of deconvolution must be almost concave!");
		}
		if ( Configuration.performChecks() && !curve2.isConvex() ) {
			throw new IllegalArgumentException("Second curve of deconvolution must be convex!");
		}
		return deconvolveAlmostConcaveConvex( curve1, curve2 );
	}
	
	/**
	 * Returns the deconvolution of an (almost) concave curve and
	 * a convex curve.
	 * @param curve1 the (almost) concave curve.
	 * @param curve2 the convex curve.
	 * @return the convolved curve.
	 */
	private static Curve deconvolveAlmostConcaveConvex( Curve curve1, Curve curve2 ) {
		int i_this  = 0;
		int i_other = 0;

		// advance this_iterator to concave part
		while(i_this < curve1.getSegmentCount()) {
			if ( curve1.getSegment( i_this ).grad.greater( new Rational( 0.0 ) ) ) {
				break;
			}
			i_this++;
		}
		// advance other_iterator to first IP after the start of this's concave part
		while(i_other < curve2.getSegmentCount()) {
			if ( curve2.getSegment( i_other ).x0.ge( curve1.getSegment( i_this ).x0 ) ) {
				break;
			}
			i_other++;
		}

		// from here on, search both curve's IPs until grad(other)>=grad(this)
		Rational kj = Rational.getPosInfinity();
		while(i_this < curve1.getSegmentCount() || i_other < curve2.getSegmentCount()) {
			Rational x_this = ( i_this < curve1.getSegmentCount() ) ?
					curve1.getSegment( i_this ).x0 : Rational.getPosInfinity();
			Rational x_other = (i_other < curve2.getSegmentCount()) ?
					curve2.getSegment(i_other).x0 : Rational.getPosInfinity();

			if ( x_other.le( x_this ) ) {
				if ( curve2.getSegment(i_other).grad.ge
						( curve1.getGradientLimitRight( curve2.getSegment( i_other ).x0) ) ) {
					kj = curve2.getSegment(i_other).x0;
					break;
				}
				i_other++;
			} else {
				if ( curve2.getGradientLimitRight( curve1.getSegment( i_this ).x0 ).ge
						( curve1.getSegment( i_this ).grad ) ) {
					kj = curve1.getSegment(i_this).x0;
					break;
				}
				i_this++;
			}
		}

		if ( kj.less( Rational.getPosInfinity() ) ) {
			Curve result = curve1.shiftLeftClipping( kj ).add( Rational.negate( curve2.f( kj ) ) );
			result.beautify();
			if ( Configuration.createLabels()
					&& (curve1.getLabel() != null || curve1.getAlias() != null
							|| curve2.getLabel() != null || curve2.getAlias() != null)) {
				String label1 = (curve1.getAlias().length() > 0) ? curve1.getAlias() : "?";
				String label2 = (curve2.getAlias().length() > 0) ? curve2.getAlias() : "?";
				result.setLabel ( "(" + label1 + "\u2298" + label2 + ")" );
			}
			return result;
		} else {
			Curve c = new Curve(1);
			c.getSegment(0).y0 = Rational.getPosInfinity();
			return c;
		}
	}
}
