/*
 *  (c) 2008 Andreas Kiefer, disco | Distributed Computer Systems Lab
 *                                   University of Kaiserslautern, Germany
 *      All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.misc.Rational;
import unikl.disco.misc.SetUtils;

import unikl.disco.curves.Curve;

import unikl.disco.nc.Configuration.Flag;
import unikl.disco.network.Edge;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Vertex;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 */
public class SeparateFlowAnalysis extends Analysis
{
	Map<Vertex, Curve> server_left_over_beta_map = null;
	Map<Vertex, Curve> server_cross_traffic_alpha_map = null;
	private Curve beta_endtoend = Curve.createZeroDelayBurst();

	@SuppressWarnings("unused")
	private SeparateFlowAnalysis(){}
	
	public SeparateFlowAnalysis( Network network ) {
		super.network = network;
	}
	/**
	 * Performs a separated flow analysis for the <code>flow_of_interest</code>
	 * and returns the end-to-end service curve.<br/>
	 * This analysis first blends out the flow of interest and then computes for
	 * each server along this flow's path the effective service curve that
	 * results if all remaining flows crossing this server receive their maximum
	 * amount of service. Then all effective service curves are concatenated to
	 * receive the end-to-end service curve from the perspective of the flow of
	 * interest.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the end-to-end service curve shall be
	 *            computed
	 * @return the end-to-end service curve
	 */
	@Override
	public boolean performAnalysis( Flow flow_of_interest )
	{
		server_left_over_beta_map = new HashMap<Vertex, Curve>();
		server_cross_traffic_alpha_map = new HashMap<Vertex, Curve>();
		beta_endtoend = Curve.createZeroDelayBurst();
		
		try{
			super.flow_of_interest = flow_of_interest;
	
			Set<Flow> flow_of_interest_as_set = new HashSet<Flow>( Collections.singleton( flow_of_interest ) );
	
			List<Edge> path = flow_of_interest.getPathInclExplSnk();
			for ( Edge e : path )
			{
				Vertex server = e.getSource();
			
				Set<Flow> flows_at_server = network.getSourceFlowSet( server );				
				Set<Flow> flows_cross_traffic = SetUtils.getDifference( flows_at_server, flow_of_interest_as_set );
				
				Curve sum_alpha_cross_traffic = network.getAggregateSourceFlowAC( server, flows_cross_traffic );

				for ( Vertex predecessor : network.getPredecessors( server ) )
				{
					flows_at_server = network.getIncomingFlows( predecessor, server );			
					flows_cross_traffic = SetUtils.getDifference( flows_at_server, flow_of_interest_as_set );
					
					sum_alpha_cross_traffic = Curve.add( sum_alpha_cross_traffic, computeOutputBound( predecessor, flows_cross_traffic ) );
				}
				
				Curve beta_effective = server.getServiceCurve();
				if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_OFF || 
						( Configuration.getUseArbMuxFlag() == Flag.NODELOCAL_FLAG & server.useFifoMultiplexing() ) )
				{
					beta_effective = computeFifoServiceCurve( sum_alpha_cross_traffic, beta_effective );
				}
				else
				{
					beta_effective = Curve.boundAtXAxis( Curve.sub( beta_effective, sum_alpha_cross_traffic ) );				
				}

				server_left_over_beta_map.put( server, beta_effective );
				server_cross_traffic_alpha_map.put( server, sum_alpha_cross_traffic );
				
				beta_endtoend = MinPlus.convolve( beta_endtoend, beta_effective );
			}
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	public Curve getEndToEndServiceCurve () {
		return beta_endtoend;
	}
	
	@Override
	public Rational getBacklogBound() {
		return Rational.max( new Rational( 0.0 ),
						Curve.getVerticalDeviation( flow_of_interest.getArrivalCurve(), beta_endtoend ));
	}
	
	@Override
	public Rational getDelayBound() {
		return Rational.max( new Rational( 0.0 ), 
						Curve.getHorizontalDeviation( flow_of_interest.getArrivalCurve(), beta_endtoend ));
	}

	public Map<Vertex, Curve> getServerLeftOverBetaMap(){
		return server_left_over_beta_map;
	}

	public Map<Vertex, Curve> getServerCrossTrafficAlphaMap(){
		return server_cross_traffic_alpha_map;
	}
}
