/*
 *  (c) 2008 Andreas Kiefer, disco | Distributed Computer Systems Lab
 *                                   University of Kaiserslautern, Germany
 *      All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.HashMap;
import java.util.Set;
import java.util.List;
import java.util.Map;

import unikl.disco.curves.Curve;
import unikl.disco.misc.Rational;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.Flag;
import unikl.disco.network.Edge;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Vertex;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 */
public class TotalFlowAnalysis extends Analysis
{
	Map<Vertex, Rational> server_backlog_bound_map = null;
	Map<Vertex, Rational> server_delay_bound_map = null;
	Map<Vertex, Curve> server_alpha_map = null;
	
	Rational total_delay = new Rational( 0.0 );
	
	@SuppressWarnings("unused")
	private TotalFlowAnalysis(){}
	
	public TotalFlowAnalysis( Network network ) {
		super.network = network;
	}
	
	/**
	 * Performs a total flow analysis for the <code>flow_of_interest</code> and
	 * returns the end-to-end delay bound. If a <code>Map</code> instance is
	 * passed via the parameter <code>server_to_backlog_bound_map</code>, the
	 * map is filled with key-value-pairs mapping servers (instances of
	 * <code>Vertex</code>) to backlog bounds (instances of <code>Double</code>
	 * ). If no such mapping is needed, pass <code>null</code> instead.<br/>
	 * A total flow analysis iterates over the path of the flow of interest,
	 * summing up the additional delay incurred by each server it traverses.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the bounds shall be computed
	 * @param server_to_backlog_bound_map
	 *            a <code>Map</code> instance that on return is filled with
	 *            &lt;server,backlog bound&gt;-pairs; pass <code>null</code> if
	 *            no such mapping is needed
	 * @return the end-to-end delay bound
	 */
	public boolean performAnalysis(	Flow flow_of_interest )
	{
		server_backlog_bound_map = new HashMap<Vertex, Rational>();
		server_delay_bound_map = new HashMap<Vertex, Rational>();
		server_alpha_map = new HashMap<Vertex, Curve>();
		
		try{
			total_delay = new Rational( 0.0 );

			List<Edge> path = flow_of_interest.getPathInclExplSnk();
			for ( Edge e : path )
			{
				Vertex server = e.getSource();
				
				Set<Flow> incoming_flows = network.getSourceFlowSet( server );
				Set<Flow> included_flows = SetUtils.getIntersection( incoming_flows, network.getEdgeFlowSet( e ) );
				Set<Flow> excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );

				Curve sum_alpha = network.getAggregateSourceFlowAC( server, incoming_flows );
				Curve sum_alpha_excl = network.getAggregateSourceFlowAC( server, excluded_flows );
				
				for ( Vertex predecessor : network.getPredecessors( server ) )
				{
					incoming_flows = network.getIncomingFlows( predecessor, server );
					included_flows = SetUtils.getIntersection( incoming_flows, network.getEdgeFlowSet( network.getEdge( predecessor, server ) ) );
					excluded_flows = SetUtils.getDifference( incoming_flows, included_flows );
	
					sum_alpha = Curve.add( sum_alpha, computeOutputBound( predecessor, included_flows ) );
					sum_alpha_excl = Curve.add( sum_alpha_excl, computeOutputBound(	predecessor, excluded_flows ) );
				}
	
				Curve beta_effective = Curve.boundAtXAxis( Curve.sub( server.getServiceCurve(), sum_alpha_excl ) );
				Rational local_delay_bound = new Rational( 0.0 );

				if( Configuration.getUseArbMuxFlag() == Flag.GLOBALLY_ON ||
					( Configuration.getUseArbMuxFlag() == Flag.NODELOCAL_FLAG & server.useArbitraryMultiplexing() ) )
				{
					local_delay_bound = Curve.getXIntersection( sum_alpha, beta_effective );				
				} else {
					local_delay_bound = Rational.max( new Rational( 0.0 ), Curve.getHorizontalDeviation( sum_alpha, beta_effective ) );	
				}

				server_delay_bound_map.put( server, local_delay_bound );
				total_delay = Rational.add( total_delay, local_delay_bound );
				server_alpha_map.put( server, sum_alpha );
								
				server_backlog_bound_map.put( server, Curve.getVerticalDeviation( sum_alpha, beta_effective ).copy() );
			}
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	@Override
	public Rational getDelayBound() {
		return total_delay;
	}

	public Map<Vertex, Rational> getServerDelayBoundMap(){
		return server_delay_bound_map;
	}
	
	@Override
	public Rational getBacklogBound() {
		Rational result = new Rational( 0.0 );
		
		for( Rational value : server_backlog_bound_map.values() ) {
			if( value.ge( result ) ) {
				result = value;
			}
		}
		
		return result.copy();
	}
	
	public Map<Vertex, Rational> getServerBacklogBoundMap(){
		return server_backlog_bound_map;
	}

	public Map<Vertex, Curve> getServerAlphaMap(){
		return server_alpha_map;
	}
}
