/*
 *  (c) 2011 Steffen Bondorf, disco | Distributed Computer Systems Lab
 *                                  University of Kaiserslautern, Germany
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.network;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration;
import unikl.disco.curves.Curve;

/**
 * 
 * @author Steffen Bondorf
 *
 */

//TODO throw more meaningful exceptions

public class Network {
	private Set<Vertex> vertices;
	private Set<Edge> edges;
	private Set<Flow> flows;

	private Vertex explicit_sink;
	
	private Map<Vertex,Set<Edge>> vertex_inEdge_map;
	private Map<Vertex,Set<Edge>> vertex_outEdge_map;

	private Map<Edge,Set<Flow>> edge_flow_map;
	private Map<Vertex,Set<Flow>> vertex_source_flow_map;
	
	public Network() {
		vertices = new HashSet<Vertex>();
		edges = new HashSet<Edge>();
		flows = new HashSet<Flow>();
		
		vertex_inEdge_map = new HashMap<Vertex,Set<Edge>>();
		vertex_outEdge_map = new HashMap<Vertex,Set<Edge>>();

		vertex_source_flow_map = new HashMap<Vertex,Set<Flow>>();
		edge_flow_map = new HashMap<Edge,Set<Flow>>();

		explicit_sink = addDummyVertex();
	}

	// ---------------------------------------------------------------------
	// vertices
	// ---------------------------------------------------------------------
	public Vertex addVertex( Curve service_curve ) {
		Vertex v = new Vertex( service_curve.copy() );
		
		vertex_inEdge_map.put( v, new HashSet<Edge>() );
		vertex_outEdge_map.put( v, new HashSet<Edge>() );

		vertices.add( v );
		
		return v;
	}
	
	public Vertex addVertex( Curve service_curve, Curve max_service_curve ) {
		Vertex v = new Vertex( service_curve.copy(), max_service_curve.copy() );
		
		vertex_inEdge_map.put( v, new HashSet<Edge>() );
		vertex_outEdge_map.put( v, new HashSet<Edge>() );

		vertices.add( v );
		
		return v;
	}
	
	private Vertex addDummyVertex() {
		Vertex v = new Vertex();
		
		vertex_inEdge_map.put( v, new HashSet<Edge>() );
		vertex_outEdge_map.put( v, new HashSet<Edge>() );

		vertices.add( v );
		
		return v;
	}

	public boolean removeVertex( Vertex v ) {
		boolean result = true;
		
		for( Edge e : getIncidentEdges( v ) ) {
			for (Flow f: edge_flow_map.get( e ) ){
				removeFlow( f );
			}
			result = result & removeEdge( e );
		}
		vertex_inEdge_map.remove( v );
		vertex_outEdge_map.remove( v );

		// All flows should already be deleted
		for ( Flow f : vertex_source_flow_map.get( v ) ){
			result = result & removeFlow( f );
		}
		result = result & vertices.remove( v );
		vertex_source_flow_map.remove( v );
		
		return result;
	}
	
	protected void addSourceFlow( Vertex source, Flow flow ){
		Set<Flow> sfs = getSourceFlowSet( source );
		
		if ( sfs.isEmpty() ) {
			sfs = new HashSet<Flow>();
			vertex_source_flow_map.put( source, sfs );
		}
		
		sfs.add( flow );
	}
	
	public Set<Flow> getSourceFlowSet( Vertex source ) {
		Set<Flow> sfs = vertex_source_flow_map.get( source );
		if ( sfs == null ){
			return new HashSet<Flow>();
		} else {
			return sfs;
		}
	}
	
	public Set<Vertex> getVertices() {
		return vertices;
	}
	
	public int numVertices() {
		return vertices.size();
	}

	public int degree( Vertex v ) {
		return inDegree(v) + outDegree(v);
	}

	public int inDegree( Vertex v ) {
		return getInEdges(v).size();
	}

	public int outDegree( Vertex v ) {
		return getOutEdges(v).size();
	}

	public Set<Edge> getInEdges( Vertex v ) {
		return vertex_inEdge_map.get(v);
	}

	public Set<Edge> getOutEdges( Vertex v ) {
		return vertex_outEdge_map.get(v);
	}

	public Set<Edge> getIncidentEdges( Vertex v ) {
		return SetUtils.getUnion( getInEdges(v), getOutEdges(v) );
	}

	public Set<Vertex> getNeighbors( Vertex v ) {
		return SetUtils.getUnion( getSuccessors(v), getPredecessors(v) );
	}

	public Set<Vertex> getSuccessors( Vertex v ) {
		Set<Vertex> successors = new HashSet<Vertex>();
		for( Edge e : getOutEdges(v) ) {
			successors.add( e.getDest() );
		}
		return successors;
	}

	public Set<Vertex> getPredecessors( Vertex v ) {
		Set<Vertex> predecessors = new HashSet<Vertex>();
		for( Edge e : getInEdges(v) ) {
			predecessors.add( e.getSource() );
		}
		return predecessors;
	}

	// -----------------------------------------------------------------------------------------
	// edges
	// -----------------------------------------------------------------------------------------
	public Set<Edge> getEdges() {
		return edges;
	}
	
	public Edge getEdge( Vertex source, Vertex sink ) {
		for( Edge e : getInEdges( sink ) ){
			if( e.getSource() == source ){
				return e;
			}
		}
		
		for( Edge e : getOutEdges( source ) ){
			if( e.getDest() == sink ){
				return e;
			}
		}
		
		return null;
	}
	
	// TODO make somehow explicit that an edge from source to destination was already in the network
	//		and thus this edge was returned instead of a newly created one
	public Edge addEdge( Vertex source, Vertex destination ) throws Exception {
		if( !vertices.contains( source ) || !vertices.contains( destination ) ){
			throw new Exception();
		}
		
		Edge existing_edge = findEdge( source, destination );
		if( existing_edge == null ) {
			
			Edge new_edge = new Edge( source, destination );
			
			edge_flow_map.put( new_edge, new HashSet<Flow>() );
			
			vertex_inEdge_map.get( new_edge.getDest() ).add( new_edge );
			vertex_outEdge_map.get( new_edge.getSource() ).add( new_edge );
			
			edges.add( new_edge );
			return new_edge;
		}else{
			return existing_edge;
		}
	}

	public boolean removeEdge( Edge e ) {
		if( e == null ){ return false; }
		
		boolean result = true;
		for( Flow f : edge_flow_map.get(e) ) {
			result = result & removeFlow( f );
		}

		// remove does not return an boolean value
		edge_flow_map.remove(e);
		
		for( Set<Edge> edges : vertex_inEdge_map.values() ) {
			edges.remove(e);
		}
		for( Set<Edge> edges : vertex_outEdge_map.values() ) {
			edges.remove(e);
		}
		
		result = result & edges.remove(e);
		
		return result;
	}

	public Edge findEdge( Vertex src, Vertex dest ) {
		Set<Edge> connecting_edge_set = SetUtils.getIntersection( getInEdges( dest ), getOutEdges( src ) );
		if( connecting_edge_set.isEmpty() ) {
			return null;
		}else{
			if( connecting_edge_set.size() > 1 ) {
				return null;
			}else{
				return connecting_edge_set.iterator().next();
			}
		}
	}
	
	/**
	 * Returns a set containing all flows traversing the given edge
	 * <code>e</code>.
	 * 
	 * @param e the traversed edge
	 * @return a set of traversing flows
	 */
	public Set<Flow> getEdgeFlowSet( Edge e ) {
		return edge_flow_map.get(e);
	}
	
	// -----------------------------------------------------------------------------------------
	// flows
	// It is not allowed to change the path of a flow,
	// i.e., the user needs to remove the old flow and insert a new one with the modified path
	// -----------------------------------------------------------------------------------------
	public Set<Flow> getFlows() {
		return flows;
	}
	
	public Flow addFlow( Curve arrival_curve, List<Edge> path ) throws Exception {
		try {
			return addFlow( arrival_curve, path, true );
		} catch (Exception e) {
			throw e;
		}
	}
	
	/**
	 * Adds a flow to the network. The flow instance has to be fully
	 * initialized, including the flow's path.
	 * 
	 * @param flow the flow to be added
	 */
	public Flow addFlow( Curve arrival_curve, Vertex source, Vertex sink ) throws Exception {
		List<Edge> path = this.getShortestPath( source, sink );
		
		if( source == sink ) {
			return addFlow( arrival_curve, sink );
		}
		
		try {
			return addFlow( arrival_curve, path, false );
		} catch (Exception e) {
			throw e;
		}
	}
	
	private Flow addFlow( Curve arrival_curve, List<Edge> path, boolean check_path ) throws Exception {
		// TODO check order of edges and reorder of necessary/possible
		// TODO check given path for possible cyclic dependencies?
		
		if( Configuration.useCheckFlowPath() && !edges.containsAll( path ) ) {
			throw new Exception();
		} else {
			Edge sink_edge = addEdge( path.get( path.size() -1 ).getDest(), explicit_sink );

			path.add( sink_edge );
			Flow f = new Flow( arrival_curve.copy(), path );

			flows.add( f );
			addSourceFlow( path.get(0).getSource(), f );
			
			for ( Edge e : path )
			{
				edge_flow_map.get( e ).add( f );
			}
			
			return f;	
		}
	}
	
	public Flow addFlow( Curve arrival_curve, Vertex single_hop ) throws Exception {
		Edge sink_edge = addEdge( single_hop, explicit_sink );
		
		List<Edge> path = new LinkedList<Edge>();
		path.add( sink_edge );
		Flow f = new Flow( arrival_curve.copy(), path );
		
		flows.add( f );
		addSourceFlow( single_hop, f );
		
		for ( Iterator<Edge> iter = path.iterator(); iter.hasNext(); )
		{
			edge_flow_map.get( iter.next() ).add( f );
		}		
		
		return f;
	}
	
	/**
	 * Removes a flow from the network.
	 * 
	 * @param flow the flow to be removed
	 * @result false if something went wrong and the network may be in an inconsistent state
	 */
	public boolean removeFlow( Flow flow )
	{
		boolean result = true;
		for ( Edge e : flow.getPathInclExplSnk() )
		{
			// single &: both sides will always be evaluated
			result = result & edge_flow_map.get( e ).remove( flow );
		}

		Set<Flow> sflows = vertex_source_flow_map.get( flow.getSource() );
		result = result & sflows.remove( flow );
		
		result = result & flows.remove( flow );
		return result;
	}

	/**
	 * Returns a set containing all flows traversing the edge from
	 * <code>predecessor</code> to <code>server</code>.
	 * 
	 * @param predecessor
	 * @param server
	 * @return a set of incoming flows
	 */
	public Set<Flow> getIncomingFlows( Vertex predecessor, Vertex server )
	{
		return edge_flow_map.get( findEdge( predecessor, server ) );
	}

	/**
	 * Returns the set of all flows that enter <code>server</code> from one of
	 * its predecessors.
	 * 
	 * @param server
	 * @return a set of all incoming flows
	 */
	public Set<Flow> getAllIncomingFlows( Vertex server )
	{
		HashSet<Flow> all_flows = new HashSet<Flow> (getSourceFlowSet( server ) );
		for ( Edge e : getInEdges( server ) )
		{
			all_flows.addAll( edge_flow_map.get( e ) );
		}
		return all_flows;
	}

	/**
	 * Returns an aggregate arrival curve for all flows originating in
	 * <code>source</code> and contained in the set <code>outgoing_flows</code>.
	 * 
	 * Returns a null curve if the intersection of those sets of curves is empty.
	 * 
	 * @param source
	 *            the source of all flows to be aggregated
	 * @param outgoing_flows
	 *            the set of all flows to be aggregated
	 * @return an aggregate arrival curve
	 */
	public Curve getAggregateSourceFlowAC( Vertex source, Set<Flow> outgoing_flows )
	{
		Curve a_out = Curve.createNullCurve();
		for ( Flow f : SetUtils.getIntersection( vertex_source_flow_map.get( source ), outgoing_flows ) )
		{
			a_out = Curve.add( a_out, f.getArrivalCurve() );
		}
		return a_out;
	}

	/**
	 * Returns the predecessor server from which a given <code>server</code>
	 * receives the flow <code>f</code>.
	 * 
	 * @param server
	 *            the receiving server
	 * @param f
	 *            the incoming flow
	 * @return the predecessor
	 */
	public Vertex getPredecessor( Vertex server, Flow f )
	{
		for ( Edge e : getInEdges(server) )
		{
			if ( edge_flow_map.get(e).contains( f ) )
			{
				return e.getSource();
			}
		}
		return null;
	}
 	
 	/**
 	 * Aggregates a set of flows by removing them from the internal management
 	 * structures and replacing it with a single dummy aggregate flow.
 	 * 
 	 * @param aggregateable_flows
 	 *            the flows to aggregate
 	 * @param joining_flows
 	 *            the set of flows with the same ingress as the aggregate
 	 * @param all_interfering_flows
 	 *            the set of all interfering flows
 	 * @param flow_egress_map
 	 *            a map of flows to egress servers
 	 * @return the aggregate flow.
 	 */
 	public Flow aggregateFlows( Set<Flow> aggregateable_flows,
 								Set<Flow> joining_flows,
 								Set<Flow> all_interfering_flows,
 								Map<Flow, Integer> flow_egress_map )
 	{
 		Flow aggregate_flow;

 		// backup egress server of the aggregateable flows
 		Integer egress_server = flow_egress_map.get( aggregateable_flows.iterator().next() );

 		// remove aggregateable flows from all local data structures...
 		joining_flows.removeAll( aggregateable_flows );
 		all_interfering_flows.removeAll( aggregateable_flows );
 		for ( Flow f : aggregateable_flows )
 		{
 			flow_egress_map.remove( f );
 		}

 		// ...and create a new aggregate flow instead.
 		aggregate_flow = Flow.createDummyFlow();
 		joining_flows.add( aggregate_flow );
 		all_interfering_flows.add( aggregate_flow );
 		flow_egress_map.put( aggregate_flow, egress_server );

 		return aggregate_flow;
 	}

 	/**
 	 * Creates a list containing the joining flows sets for each server along
 	 * the path. Does not include the flows_of_interest.
 	 * 
 	 * @param path
 	 * @param flows_of_interest
 	 * @return a list containing sets of joining flows
 	 */
 	public List<Set<Flow>> getJoiningFlowLists(List<Edge> path, Set<Flow> flows_of_interest )
 	{
 		Set<Flow> present_flows = new HashSet<Flow>( flows_of_interest );
		List<Set<Flow>> joining_flow_sets = new ArrayList<Set<Flow>>();
 		for ( Edge e : path )
 		{
 			Vertex server = e.getSource();

 			Set<Flow> joining_flows = SetUtils.getDifference( getAllIncomingFlows( server ), present_flows );
 			joining_flow_sets.add( joining_flows );

 			present_flows.addAll( joining_flows );

 			Set<Flow> leaving_flows = new HashSet<Flow>();
 			for ( Edge out_e : getOutEdges(server) )
 			{
 				if ( out_e == e )
 				{
 					// skip outgoing edge which is on path
 					continue;
 				}
 				leaving_flows.addAll( edge_flow_map.get(out_e) );
 			}
 			present_flows = SetUtils.getDifference( present_flows, leaving_flows );
 		}
 		return joining_flow_sets;
 	}

 	/**
 	 * Creates a list containing a set of rejoining flows for each server along
 	 * the path. Rejoining flows are flows that have joined the flow of interest
 	 * before, then took a different path and finally join the flow of interest
 	 * again.
 	 * 
 	 * @param joining_flow_sets
 	 *            the list of joining flows
 	 * @return a list containing sets of joining flows
 	 */
 	public List<Set<Flow>> getRejoiningFlowLists( List<Set<Flow>> joining_flow_sets )
 	{
 		Set<Flow> previously_joined_flows = new HashSet<Flow>();
 		List<Set<Flow>> rejoining_flow_lists = new LinkedList<Set<Flow>>();
 		for ( int i = 0; i < joining_flow_sets.size(); i++ )
 		{
 			Set<Flow> rejoining_flows = SetUtils.getIntersection( previously_joined_flows, joining_flow_sets.get( i ) );
 			rejoining_flow_lists.add( rejoining_flows );
 			previously_joined_flows.addAll( joining_flow_sets.get( i ) );
 		}
 		return rejoining_flow_lists;
 	}

 	/**
 	 * Computes the map of flows to egress point (an Integer index into the list
 	 * of servers on the path).
 	 * 
 	 * @param path
 	 * @param all_cross_traffic_flows
 	 * @return the map
 	 */
 	public Map<Flow, Integer> getFlowEgressMap(	List<Edge> path, Set<Flow> all_cross_traffic_flows )
 	{
 		Map<Flow, Integer> map = new HashMap<Flow, Integer>();
 		int i = -1;
 		for ( Edge e : path )
 		{
 			i++;
 			Vertex server = e.getSource();
 			Set<Flow> outgoing_flows = new HashSet<Flow>();
 			for ( Edge out_e : getOutEdges( server ) )
 			{
 				if ( out_e == e )
 				{
 					// skip outgoing edge which is on path
 					continue;
 				}
 				outgoing_flows.addAll( edge_flow_map.get(out_e) );
 			}
 			outgoing_flows = SetUtils.getIntersection( outgoing_flows, all_cross_traffic_flows );
 			for ( Flow f : outgoing_flows )
 			{
 				if ( !map.containsKey( f ) )
 				{
 					map.put( f, Integer.valueOf( i ) ); // store only the first
 														// egress point of a flow
 				}
 			}
 		}
 		for ( Flow f : all_cross_traffic_flows )
 		{
 			if ( !map.containsKey( f ) )
 			{
 				map.put( f, Integer.valueOf( path.size() - 1 ) );
 			}
 		}
 		return map;
 	}

 	/**
 	 * Creates a partition of the set of flows joining this server so that each
 	 * partition contains flows with identical egress point.
 	 * 
 	 * @param path
 	 * @param server
 	 *            the ingress server
 	 * @param joining_flows
 	 *            the set of flows joining at the ingress server
 	 * @param flow_egress_map
 	 *            the flow-to-egress-point map
 	 * @return a set of flow sets with same egress point
 	 */
 	public Set<Set<Flow>> getAggregateableFlowSets(	List<Edge> path,
 														int server,
 														Set<Flow> joining_flows,
 														Map<Flow, Integer> flow_egress_map )
 	{
 		Set<Set<Flow>> aggregateable_flows = new HashSet<Set<Flow>>();
 		List<Flow> remaining_flows = new LinkedList<Flow>( joining_flows );

 		while ( remaining_flows.size() > 0 )
 		{
 			Flow flow = remaining_flows.remove( remaining_flows.size() - 1 );

 			Integer common_egress_server = flow_egress_map.get( flow );
 			if ( common_egress_server == null || common_egress_server.intValue() < server )
 			{ // i.e. this flow is rejoining
 				common_egress_server = Integer.valueOf( findEgressPoint( path, server, flow ) );
 			}

 			Set<Flow> flow_set = new HashSet<Flow>( Collections.singleton( flow ) );
 			for ( int i = remaining_flows.size() - 1; i >= 0; i-- )
 			{
 				flow = remaining_flows.get( i );

 				Integer egress_server = flow_egress_map.get( flow );
 				if ( egress_server == null || egress_server.intValue() < server )
 				{ // i.e. this flow is rejoining
 					egress_server = Integer.valueOf( findEgressPoint( path, server, flow ) );
 				}

 				if ( egress_server.intValue() == common_egress_server.intValue() )
 				{
 					flow_set.add( flow );
 					remaining_flows.remove( i );
 				}
 			}
 			aggregateable_flows.add( flow_set );
 		}

 		return aggregateable_flows;
 	}
 	
	/**
	 * 
	 * @param src
	 * @param snk
	 * @return
	 */
    public List<Edge> getShortestPath( Vertex src, Vertex snk ) {
		Set<Vertex> visited = new HashSet<Vertex>();
		Map<Vertex,List<Edge>> paths = new HashMap<Vertex,List<Edge>>();
		
		paths.put( src, new LinkedList<Edge>() );
		
		LinkedList<Vertex> queue = new LinkedList<Vertex>();
		queue.add(src);
		visited.add(src);
		
		while ( !queue.isEmpty() ) {
			Vertex v = queue.getLast();
			queue.remove(v);

			for( Vertex u : getSuccessors(v) ) {

				List<Edge> path_tmp = new LinkedList<Edge>( paths.get(v) );
				path_tmp.add( findEdge( v, u ) );
				
				if( !visited.contains(u) ) {
					// add path
					paths.put( u, path_tmp );
					queue.add( u );
					visited.add( u );
				}else{
					// compare path lengths
					if( paths.get( u ).size() > path_tmp.size() ) {
						paths.put( u, path_tmp );
						// TODO is it necessary to recheck the path to u?
						queue.add( u );
					}
				}
			}
		}
		return paths.get( snk );
	}

 	/**
 	 * Returns the index of the server at which the given flow exits the path.
 	 * This point must be after the server specified by server_idx.
 	 * 
 	 * @param path
 	 *            the path
 	 * @param server_idx
 	 *            the index of the server where the flow joins the path
 	 * @param flow
 	 *            the flow
 	 * @return the index of the server in the path
 	 */
 	public int findEgressPoint( List<Edge> path, int server_idx, Flow flow )
 	{
 		int i = -1;
 		for ( Edge e : path )
 		{
 			i++;

 			if ( i < server_idx )
 			{
 				continue;
 			}

 			Vertex server = e.getSource();
 			for ( Edge out_e : getOutEdges(server) )
 			{
 				if ( out_e == e )
 				{
 					// skip outgoing edge which is on path
 					continue;
 				}
 				if ( edge_flow_map.get(out_e).contains( flow ) )
 				{
 					return i;
 				}
 			}
 		}
 		return path.size() - 1;
 	}

 	/**
 	 * Creates a fresh flow substitute for a rejoining flow and updates the
 	 * respective management information.
 	 * 
 	 * @param rejoining_flows_in_aggregate
 	 * @param egress_server
 	 * @param aggregateable_flows
 	 * @param all_interfering_flows
 	 * @param joining_flows
 	 * @param flow_egress_map
 	 */
 	public void createSubstituteFlows(	Set<Flow> rejoining_flows_in_aggregate,
 											int egress_server,
 											Set<Flow> aggregateable_flows,
 											Set<Flow> all_interfering_flows,
 											Set<Flow> joining_flows,
 											Map<Flow, Integer> flow_egress_map )
 	{
 		for ( Flow flow : rejoining_flows_in_aggregate )
 		{
 			Flow substitute = Flow.createDummyFlow();
 			aggregateable_flows.remove( flow );
 			aggregateable_flows.add( substitute );
 			all_interfering_flows.add( substitute );
 			joining_flows.remove( flow );
 			joining_flows.add( substitute );
 			flow_egress_map.put( substitute, Integer.valueOf( egress_server ) );
 		}
 	}

 	/**
 	 * Returns the server at which the flows in <code>flows_of_interest</code>
 	 * first all meet each other (when viewed from the source). When viewed from
 	 * <code>server</code> towards the sink, this is the last server where all
 	 * flows are still together.
 	 * 
 	 * @param server
 	 * @param flows_of_interest
 	 * @return the splitting point server
 	 */
 	public Vertex findSplittingPoint( Vertex server, Set<Flow> flows_of_interest )
 	{
 		Flow f = flows_of_interest.iterator().next();

 		List<Edge> path = f.getSubPath( f.getSource(), server );

 		for ( Edge e : path )
 		{
 			Set<Flow> flow_list = edge_flow_map.get(e);
 			if ( SetUtils.getIntersection( flow_list, flows_of_interest ).size() == flows_of_interest.size() )
 			{
 				return e.getSource();
 			}
 		}
 		return server;
 	}
}
