package unikl.disco.tests;

import static org.junit.Assert.assertEquals;
import org.junit.BeforeClass;
import org.junit.Test;

import unikl.disco.curves.Curve;
import unikl.disco.misc.Rational;
import unikl.disco.nc.Configuration;
import unikl.disco.nc.Configuration.Flag;
import unikl.disco.nc.TotalFlowAnalysis;
import unikl.disco.nc.SeparateFlowAnalysis;
import unikl.disco.nc.PmooSfaAnalysis;

import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Vertex;

public class Tandem_2SCs_2Flows_1ACs_1Path
{
	static final Curve service_curve_0 = Curve.createRateLatency( 10, 10 ); // latency, rate
	static final Curve service_curve_1 = Curve.createRateLatency( 6, 6 ); // latency, rate
	static final Curve arrival_curve = Curve.createTokenBucket( 12.5, 2.5 ); // burst, rate
	static final Curve max_service_curve = Curve.createRateLatency( 5, 15 ); // latency, rate

	static Network network;
	static Vertex v0, v1;
	static Flow f0, f1;
	
	@BeforeClass
	static public void createNetwork()
	{
		network = new Network();
		v0 = network.addVertex( service_curve_0, max_service_curve );
		v0.setUseGamma( false );
		v0.setUseExtraGamma( false );
		
		v1 = network.addVertex( service_curve_1, max_service_curve );
		v1.setUseGamma( false );
		v1.setUseExtraGamma( false );

		try {
			network.addEdge( v0, v1 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}

		try {	
			f0 = network.addFlow( arrival_curve, v0, v1 );
			f1 = network.addFlow( arrival_curve, v0, v1 );
		} catch (Exception e) {
			System.out.println( e.toString() );
			return;
		}
	}
	
//--------------------Flow 0--------------------	
	@Test
	public void f0_tfa_fifoMux()
	{
		System.out.println( "Flow of interest : " + f0.toString() );
		System.out.println();
		
		v0.setUseFifoMultiplexing( true ); // default = true
		v1.setUseFifoMultiplexing( true ); // default = true
		
		System.out.println( "--- Total Flow Analysis (FIFO_MUX) ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		if ( tf.performAnalysis( f0 ) ){
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().values().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().values().toString() );
			System.out.println();
		} else {
			System.out.println( "TFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("TFA FIFO delay", new Rational( 31 ), tf.getDelayBound() );
		assertEquals("TFA FIFO backlog", new Rational( 105 ), tf.getBacklogBound() );
	}
	
	@Test
	public void f0_tfa_arbMux()
	{
		System.out.println( "Flow of interest : " + f0.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- Total Flow Analysis (ARB_MUX) ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		if ( tf.performAnalysis( f0 ) ){
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().values().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().values().toString() );
			System.out.println();
		} else {
			System.out.println( "TFA analysis failed" );
			System.out.println();
		}

		assertEquals("TFA ARB delay", new Rational( 136 ), tf.getDelayBound() );
		assertEquals("TFA ARB backlog", new Rational( 105 ), tf.getBacklogBound() );
	}
	
	@Test
	public void f0_sfa_fifoMux()
	{
		System.out.println( "Flow of interest : " + f0.toString() );
		System.out.println();
		
		Configuration.setArbMuxFlag( Flag.GLOBALLY_OFF );
		
		System.out.println( "--- Separated Flow Analysis (FIFO_MUX) ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		if ( sf.performAnalysis( f0 ) ){
			System.out.println( "e2e SFA SC      : " + sf.getEndToEndServiceCurve() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetaMap().values().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "SFA analysis failed" );
			System.out.println();
		}

		Configuration.setArbMuxFlag( Flag.NODELOCAL_FLAG );
		
		assertEquals("SFA FIFO delay", new Rational( 379, 14 ), sf.getDelayBound() );
		assertEquals("SFA FIFO backlog", new Rational( 71.25 ), sf.getBacklogBound() );
	}
	
	@Test
	public void f0_sfa_arbMux()
	{
		System.out.println( "Flow of interest : " + f0.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- Separated Flow Analysis (ARB_MUX) ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		if ( sf.performAnalysis( f0 ) ){
			System.out.println( "e2e SFA SC      : " + sf.getEndToEndServiceCurve() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetaMap().values().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "SFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("SFA ARB delay", new Rational( 277, 7 ), sf.getDelayBound() );
		assertEquals("SFA ARB backlog", new Rational( 102.5 ), sf.getBacklogBound() );
	}
	
	@Test
	public void f0_pmoo_arbMux()
	{
		System.out.println( "Flow of interest : " + f0.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- PMOO Analysis (ARB_MUX) ---" );
		PmooSfaAnalysis pmoo = new PmooSfaAnalysis( network );
		
		if ( pmoo.performAnalysis( f0 ) ){
			System.out.println( "e2e PMOO SFA SC : " + pmoo.getEndToEndServiceCurve() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "PMOO SFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("PMOO ARB delay", new Rational( 242, 7 ), pmoo.getDelayBound() );
		assertEquals("PMOO ARB backlog", new Rational( 90 ), pmoo.getBacklogBound() );
	}

//--------------------Flow 1--------------------	
	@Test
	public void f1_tfa_fifoMux()
	{
		System.out.println( "Flow of interest : " + f1.toString() );
		System.out.println();
		
		v0.setUseFifoMultiplexing( true ); // default = true
		v1.setUseFifoMultiplexing( true ); // default = true
		
		System.out.println( "--- Total Flow Analysis (FIFO_MUX) ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		if ( tf.performAnalysis( f1 ) ){
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().values().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().values().toString() );
			System.out.println();
		} else {
			System.out.println( "TFA analysis failed" );
			System.out.println();
		}

		assertEquals("TFA FIFO delay", new Rational( 31 ), tf.getDelayBound() );
		assertEquals("TFA FIFO backlog", new Rational( 105 ), tf.getBacklogBound() );
	}
	
	@Test
	public void f1_tfa_arbMux()
	{
		System.out.println( "Flow of interest : " + f1.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- Total Flow Analysis (ARB_MUX) ---" );
		TotalFlowAnalysis tf = new TotalFlowAnalysis( network );
		
		if ( tf.performAnalysis( f1 ) ){
			System.out.println( "delay bound     : " + tf.getDelayBound() );
			System.out.println( "     per server : " + tf.getServerDelayBoundMap().values().toString() );
			System.out.println( "backlog bound   : " + tf.getBacklogBound() );
			System.out.println( "     per server : " + tf.getServerBacklogBoundMap().values().toString() );
			System.out.println();
		} else {
			System.out.println( "TFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("TFA ARB delay", new Rational( 136 ), tf.getDelayBound() );
		assertEquals("TFA ARB backlog", new Rational( 105 ), tf.getBacklogBound() );
	}
	
	@Test
	public void f1_sfa_fifoMux()
	{
		System.out.println( "Flow of interest : " + f1.toString() );
		System.out.println();
		
		Configuration.setArbMuxFlag( Flag.GLOBALLY_OFF );
		
		System.out.println( "--- Separated Flow Analysis (FIFO_MUX) ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		if ( sf.performAnalysis( f1 ) ){
			System.out.println( "e2e SFA SC      : " + sf.getEndToEndServiceCurve() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetaMap().values().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "SFA analysis failed" );
			System.out.println();
		}
		
		Configuration.setArbMuxFlag( Flag.NODELOCAL_FLAG );

		assertEquals("SFA FIFO delay", new Rational( 379, 14 ), sf.getDelayBound() );
		assertEquals("SFA FIFO backlog", new Rational( 71.25 ), sf.getBacklogBound() );
	}
	
	@Test
	public void f1_sfa_arbMux()
	{
		System.out.println( "Flow of interest : " + f1.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- Separated Flow Analysis (ARB_MUX) ---" );
		SeparateFlowAnalysis sf = new SeparateFlowAnalysis( network );

		if ( sf.performAnalysis( f1 ) ){
			System.out.println( "e2e SFA SC      : " + sf.getEndToEndServiceCurve() );
			System.out.println( "     per server : " + sf.getServerLeftOverBetaMap().values().toString() );
			System.out.println( "delay bound     : " + sf.getDelayBound() );
			System.out.println( "backlog bound   : " + sf.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "SFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("SFA ARB delay", new Rational( 277, 7 ), sf.getDelayBound() );
		assertEquals("SFA ARB backlog", new Rational( 102.5 ), sf.getBacklogBound() );
	}
	
	@Test
	public void f1_pmoo_arbMux()
	{
		System.out.println( "Flow of interest : " + f1.toString() );
		System.out.println();
		
		v0.setUseArbitraryMultiplexing( true ); // default = true
		v1.setUseArbitraryMultiplexing( true ); // default = true
		
		System.out.println( "--- PMOO Analysis (ARB_MUX) ---" );
		PmooSfaAnalysis pmoo = new PmooSfaAnalysis( network );
		
		if ( pmoo.performAnalysis( f1 ) ){
			System.out.println( "e2e PMOO SFA SC : " + pmoo.getEndToEndServiceCurve() );
			System.out.println( "delay bound     : " + pmoo.getDelayBound() );
			System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
			System.out.println();
		} else {
			System.out.println( "PMOO SFA analysis failed" );
			System.out.println();
		}
		
		assertEquals("PMOO ARB delay", new Rational( 242, 7 ), pmoo.getDelayBound() );
		assertEquals("PMOO ARB backlog", new Rational( 90 ), pmoo.getBacklogBound() );
	}
}